import os
import re
import time
from copy import deepcopy
from datetime import datetime, timedelta

import dateutil.parser
from source import helper, shared, tables
from source.shared import utc
from source.tables import ColumnSpec
from urllib3.connectionpool import MaxRetryError

from seeq import sdk, spy
from seeq.base import util
from seeq.base.seeq_names import SeeqNames
from seeq.sdk import *
from seeq.sdk.rest import ApiException

SORT_BY_ITEM_GUID = ["ID"]
SORT_BY_ITEM_NAME = ["Name"]


def datasource_list(**kwargs):
    """
    List registered datasources and associated information.
    :param csv: CSV file to write items to.
    """
    client = spy.client

    _filter = ""
    _types = ["Datasource"]

    items_api = sdk.ItemsApi(client)

    tables.print_table_from_sdk_output(
        lambda offset: items_api.search_items(
            filters=[_filter, SeeqNames.API.Flags.include_unsearchable],
            types=_types,
            offset=offset,
            limit=helper.get_default_page_size(),
            order_by=SORT_BY_ITEM_NAME,
        ),
        "items",
        [
            ColumnSpec("count", "#", ">", 4),
            ColumnSpec("name", "Datasource Name", "<", 40),
            ColumnSpec(
                "",
                "Datasource Class",
                "<",
                30,
                lambda value, item: _get_item_property(
                    item.id, SeeqNames.Properties.datasource_class, items_api
                ),
            ),
            ColumnSpec(
                "",
                "Datasource ID",
                "<",
                36,
                lambda value, item: _get_item_property(
                    item.id, SeeqNames.Properties.datasource_id, items_api
                ),
            ),
            ColumnSpec(
                "",
                "Stored",
                "<",
                6,
                lambda value, item: _get_item_property(
                    item.id, SeeqNames.Properties.stored_in_seeq, items_api
                ),
            ),
            ColumnSpec(
                "",
                "Cached",
                "<",
                6,
                lambda value, item: _get_item_property(
                    item.id, SeeqNames.Properties.cache_enabled, items_api
                ),
            ),
            ColumnSpec(
                "",
                "Archived",
                "<",
                8,
                lambda value, item: _get_item_property(
                    item.id, SeeqNames.Properties.archived, items_api
                ),
            ),
            ColumnSpec("id", "ID", "<", 36),
        ],
        csv_file_name=kwargs.get("csv"),
    )


def _get_item_property(_id, property_name, items_api):
    try:
        return items_api.get_property(id=_id, property_name=property_name).value
    except ApiException:
        return ""


def get_datasource_item(client, **kwargs):
    items_api = sdk.ItemsApi(client)

    if kwargs.get("id"):
        datasource_item = items_api.get_item_and_all_properties(id=kwargs.get("id"))
        if datasource_item.type != "Datasource":
            raise RuntimeError(
                "Provided Seeq ID does not match a Datasource; please try again with the ID of a "
                "Datasource.\n"
                'Execute "datasource.datasource_list()" to find the ID you want and then re-run this command.'
            )

    else:
        _filters = list()
        if kwargs.get("datasource_class"):
            _filters.append("Datasource Class==" + kwargs.get("datasource_class"))
        if kwargs.get("datasource_id"):
            _filters.append("Datasource ID==" + kwargs.get("datasource_id"))
        if kwargs.get("datasource_name"):
            _filters.append("Name==" + kwargs.get("datasource_name"))

        filters_arg = [" && ".join(_filters), SeeqNames.API.Flags.include_unsearchable]

        item_search_list = items_api.search_items(
            types=["Datasource"],
            filters=filters_arg,
            limit=helper.get_default_page_size(),
        )  # type: ItemSearchPreviewPaginatedListV1
        if len(item_search_list.items) == 0:
            raise RuntimeError("Datasource not found")
        if len(item_search_list.items) > 1:
            tables.print_table_from_sdk_output(
                lambda offset: items_api.search_items(
                    filters=filters_arg,
                    offset=offset,
                    limit=helper.get_default_page_size(),
                    order_by=SORT_BY_ITEM_GUID,
                ),
                "items",
                [
                    ColumnSpec("name", "Datasource Name", "<", 30),
                    ColumnSpec(
                        "",
                        "Datasource Class",
                        "<",
                        25,
                        transform=lambda value, item: _get_item_property(
                            item.id, SeeqNames.Properties.datasource_class, items_api
                        ),
                    ),
                    ColumnSpec(
                        "",
                        "Datasource ID",
                        "<",
                        36,
                        transform=lambda value, item: _get_item_property(
                            item.id, SeeqNames.Properties.datasource_id, items_api
                        ),
                    ),
                ],
            )

            raise RuntimeError(
                "Multiple datasources found, please narrow your search.\n"
                'Execute "datasource.datasource_list()" to find the ID you want and then re-run this command\n'
                "with the --id flag."
            )

        datasource_item = item_search_list.items[0]

    datasource_id = items_api.get_property(
        id=datasource_item.id, property_name=SeeqNames.Properties.datasource_id
    ).value
    datasource_class = items_api.get_property(
        id=datasource_item.id, property_name=SeeqNames.Properties.datasource_class
    ).value
    datasource_name = items_api.get_property(
        id=datasource_item.id, property_name=SeeqNames.Properties.name
    ).value

    return shared.Bunch(
        id=datasource_item.id,
        name=datasource_name,
        datasource_class=datasource_class,
        datasource_id=datasource_id,
    )


def _ancestor_path(ancestors):
    if not ancestors:
        return ""

    path = ""
    for ancestor in ancestors:
        if len(path) > 0:
            path = path + " "

        path += ancestor.name + " >>"

    return path


def items(**kwargs):
    """
    List the items associated with the datasource.
    :param datasource_name: Datasource Name of datasource to search.
    :param datasource_class: Datasource Class of datasource to search.
    :param datasource_id: Datasource ID of datasource to search.
    :param id: Seeq ID of datasource to search
    :param filter: filter (e.g. "*.PV").
    :param include_archive: include archived items
    :param csv: CSV file to write items to.
    """

    client = spy.client

    items_api = sdk.ItemsApi(client)

    datasource_item = get_datasource_item(client, **kwargs)

    _filter = "Datasource Class==" + datasource_item.datasource_class
    _filter += " && Datasource ID==" + datasource_item.datasource_id

    if kwargs.get("filter"):
        _filter += " && Name ~= " + kwargs.get("filter")

    _filters = [_filter]
    if kwargs.get("include_archived"):
        _filters.append(SeeqNames.API.Flags.include_unsearchable)

    tables.print_table_from_sdk_output(
        lambda offset: items_api.search_items(
            filters=_filters,
            offset=offset,
            limit=helper.get_default_page_size(),
            order_by=SORT_BY_ITEM_NAME,
        ),
        "items",
        [
            ColumnSpec("count", "#", ">", 8),
            ColumnSpec(
                "ancestors",
                "Path",
                ">",
                40,
                lambda ancestors, item: _ancestor_path(ancestors),
            ),
            ColumnSpec("name", "Name", "<", 40),
            ColumnSpec("type", "Type", "<", 15),
            ColumnSpec("value_unit_of_measure", "UOM", "<", 8),
            ColumnSpec("id", "ID", "<", 36),
        ],
        csv_file_name=kwargs.get("csv"),
    )


def recount(**kwargs):
    """
    Recount the totals for items associated with the datasource.
    :param datasource_name: Datasource Name of datasource to search.
    :param datasource_class: Datasource Class of datasource to search.
    :param datasource_id: Datasource ID of datasource to search.
    :param id: Seeq ID of datasource to search
    :param filter: filter (e.g. "*.PV").
    :param include_archive: include archived items
    :param csv: CSV file to write items to.
    """

    client = spy.client

    items_api = sdk.ItemsApi(client)
    datasources_api = sdk.DatasourcesApi(client)

    datasource_item = get_datasource_item(client, **kwargs)

    print("Current counts:")

    _print_datasource_item_counts(items_api, datasource_item)

    print(
        "\nRe-counting items associated with this datasource. This may take several minutes depending on"
    )
    print("the number of items in the datasource.")

    datasources_api.recount(id=datasource_item.id)

    print("Finished successfully. New counts:")

    _print_datasource_item_counts(items_api, datasource_item, kwargs.get("csv"))

    print('\nTotals in the Seeq Workbench "Connections" list should be correct.')


def _print_datasource_item_counts(items_api, datasource_item, csv_file=None):
    table = tables.start_table(
        [
            ColumnSpec("item_type", "Item Type", "<", 20),
            ColumnSpec("item_count", "Item Count", ">", 10),
        ],
        csv_file,
    )

    tables.write_table_row(
        table,
        shared.Bunch(
            item_type="Signals",
            item_count=_get_property_or_default(
                items_api, datasource_item.id, "Signal Count", "0"
            ),
        ),
    )

    tables.write_table_row(
        table,
        shared.Bunch(
            item_type="Conditions",
            item_count=_get_property_or_default(
                items_api, datasource_item.id, "Condition Count", "0"
            ),
        ),
    )

    tables.write_table_row(
        table,
        shared.Bunch(
            item_type="Assets",
            item_count=_get_property_or_default(
                items_api, datasource_item.id, "Asset Count", "0"
            ),
        ),
    )

    tables.write_table_row(
        table,
        shared.Bunch(
            item_type="Scalars",
            item_count=_get_property_or_default(
                items_api, datasource_item.id, "Scalar Count", "0"
            ),
        ),
    )

    tables.write_table_row(
        table,
        shared.Bunch(
            item_type="Relationships",
            item_count=_get_property_or_default(
                items_api, datasource_item.id, "Relationship Count", "0"
            ),
        ),
    )

    tables.finish_table(table)


def _get_property_or_default(items_api, _id, property_name, default):
    try:
        prop = items_api.get_property(
            id=_id, property_name=property_name
        )  # type: PropertyOutputV1
        return prop.value
    except ApiException:
        return default


def archive(**kwargs):
    """
    Archives a datasource and associated items.
    :param datasource_name: Datasource Name of datasource to search.
    :param datasource_class: Datasource Class of datasource to search.
    :param datasource_id: Datasource ID of datasource to search.
    :param id: Seeq ID of datasource to search.
    :param filter: filter (e.g. "*.PV").
    :param restore: restore the archive status.
    :param csv: CSV file to write items to.
    """

    client = spy.client

    items_api = sdk.ItemsApi(client)

    datasource_item = get_datasource_item(client, **kwargs)

    datasource_id = items_api.get_property(
        id=datasource_item.id, property_name=SeeqNames.Properties.datasource_id
    ).value
    datasource_class = items_api.get_property(
        id=datasource_item.id, property_name=SeeqNames.Properties.datasource_class
    ).value

    _filter = "Datasource Class==" + datasource_class
    _filter += " && Datasource ID==" + datasource_id

    if kwargs.get("filter"):
        _filter += " && Name ~= " + kwargs.get("filter")

    if not kwargs.get("filter") and kwargs.get("restore"):
        # we're reverting the archive flag
        # the datasource needs to be restore before the items so that permissions are inherited correctly
        _process_archive_command_on_item(datasource_item, client, kwargs.get("restore"))
        print("\nDatasource item has been UN-ARCHIVED.")

    tables.print_table_from_sdk_output(
        lambda offset: items_api.search_items(
            filters=[_filter, SeeqNames.API.Flags.include_unsearchable],
            offset=offset,
            limit=helper.get_default_page_size(),
            order_by=SORT_BY_ITEM_GUID,
        ),
        "items",
        [
            ColumnSpec("count", "#", ">", 8),
            ColumnSpec("total_time_str", "Time", ">", 8),
            ColumnSpec(
                "ancestors",
                "Path",
                ">",
                40,
                lambda ancestors, item: _ancestor_path(ancestors),
            ),
            ColumnSpec("name", "Name", "<", 40),
            ColumnSpec("type", "Type", "<", 15),
            ColumnSpec("value_unit_of_measure", "UOM", "<", 8),
            ColumnSpec("id", "ID", "<", 36),
        ],
        # Note that we exclude the datasource item here so that we archive everything else first, then do the
        # datasource at the end.
        lambda item: (
            _process_archive_command_on_item(item, client, kwargs.get("restore"))
            if item.type != "Datasource"
            else ""
        ),
        csv_file_name=kwargs.get("csv"),
    )

    print(
        "\nAll of the above items have been %s."
        % ("UN-ARCHIVED" if kwargs.get("restore") else "ARCHIVED")
    )

    if not kwargs.get("filter") and not kwargs.get("restore"):
        # We're archiving the whole datasource
        # datasource should be archived last since we want the permissions to be inherited correctly
        _process_archive_command_on_item(datasource_item, client, kwargs.get("restore"))
        print("\nDatasource item has been ARCHIVED.")


def _process_archive_command_on_item(item, client, restore):
    items_api = sdk.ItemsApi(client)

    property_input = PropertyInputV1()
    property_input.value = True if not restore else False
    items_api.set_property(id=item.id, property_name="Archived", body=property_input)


def cache(**kwargs):
    """
    Turn persistent datasource caching on/off and clear persistent caching to force reload from datasource.
    :param command: on, off, clear
    :param datasource_name: Datasource Name of datasource to search.
    :param datasource_class: Datasource Class of datasource to search.
    :param datasource_id: Datasource ID of datasource to search.
    :param id: Seeq ID of datasource to search.
    :param filter: filter (e.g. "*.PV").
    :param input_csv: CSV file with ID or Name column for items to affect
    :param restore: restore the archive status.
    :param csv: CSV file to write items to.
    """

    if kwargs.get("command"):
        if kwargs.get("command") not in ["on", "off", "clear"]:
            print("You must provide a command: on, off, clear")
            return
        cmd = kwargs.get("command")
    else:
        print("You must provide a command: on, off, clear")
        return

    client = spy.client

    items_api = sdk.ItemsApi(client)

    datasource_item = get_datasource_item(client, **kwargs)

    _types = [
        "StoredSignal",
        "CalculatedSignal",
        "StoredCondition",
        "CalculatedCondition",
        "CalculatedScalar",
    ]
    _filter = "Datasource Class==" + datasource_item.datasource_class
    _filter += " && Datasource ID==" + datasource_item.datasource_id

    # Be as selective as possible. This means pagination is only necessary for the 'clear' command.
    # For 'clear' or 'off' we want to find items where cache is enabled; for 'on' we want those with it disabled.
    _filter += " && Cache Enabled == %s" % ("true" if cmd != "on" else "false")

    if kwargs.get("filter") and kwargs.get("input_csv"):
        raise RuntimeError("filter and input_csv arguments are mutually exclusive")

    if kwargs.get("input_csv"):
        column_specs = [
            ColumnSpec("count", "#", ">", 8),
            ColumnSpec("name", "Name", "<", 40),
            ColumnSpec("type", "Type", "<", 15, console=False),
            ColumnSpec("id", "ID", "<", 36, console=False),
        ]

        header, rows = tables.read_table(kwargs.get("input_csv"))

        id_column = tables.find_column_index(header, "ID")
        name_column = tables.find_column_index(header, "Name")

        if name_column is None and id_column is None:
            raise RuntimeError('"ID" nor "Name" column not found in input CSV file')

        table = tables.start_table(column_specs, kwargs.get("csv"))

        for row in rows:
            item_id = row[id_column] if id_column is not None else None
            item_name = row[name_column] if name_column is not None else None

            items = []
            if item_name:
                _filter += "&& Name == %s" % item_name
                search_results = items_api.search_items(
                    filters=[_filter, SeeqNames.API.Flags.include_unsearchable],
                    types=_types,
                )  # type: ItemSearchPreviewPaginatedListV1
                if len(search_results.items) > 0:
                    items = search_results.items
                else:
                    items = [shared.Bunch(name=item_name, type="", id="NOT FOUND")]
            elif item_id:
                _filter += "&& ID == %s" % item_id
                try:
                    item = items_api.get_item_and_all_properties(
                        id=item_id
                    )  # type: ItemOutputV1
                    items = [shared.Bunch(name=item.name, type=item.type, id=item.id)]
                except Exception:
                    items = [shared.Bunch(name=item_name, type="", id="NOT FOUND")]

            for _item in items:
                if _item.id != "NOT FOUND":
                    _process_cache_command_on_item(cmd, _item, client)
                tables.write_table_row(table, _item)

        tables.finish_table(table)

    else:
        column_specs = [
            ColumnSpec("count", "#", ">", 8),
            ColumnSpec("total_time_str", "Time", ">", 8),
            ColumnSpec(
                "ancestors",
                "Path",
                ">",
                40,
                lambda ancestors, _item: _ancestor_path(ancestors),
            ),
            ColumnSpec("name", "Name", "<", 40),
            ColumnSpec("type", "Type", "<", 15, console=False),
            ColumnSpec("id", "ID", "<", 36, console=False),
        ]

        if kwargs.get("filter"):
            _filter += " && Name ~= " + kwargs.get("filter")

        tables.print_table_from_sdk_output(
            lambda offset: items_api.search_items(
                filters=[_filter, SeeqNames.API.Flags.include_unsearchable],
                types=_types,
                offset=offset if cmd == "clear" else 0,  # only paginate 'clear'
                limit=helper.get_default_page_size(),
                order_by=SORT_BY_ITEM_GUID,
            ),
            "items",
            column_specs,
            lambda _item: _process_cache_command_on_item(cmd, _item, client),
            csv_file_name=kwargs.get("csv"),
        )

    if cmd == "on":
        message = "caching turned ON"
    elif cmd == "off":
        message = "caching turned OFF"
    else:
        message = "cache CLEARED"

    print("\nPersistent %s for all of the above items." % message)

    if cmd != "on":
        print("Note: Only items with caching ON were affected.")

    if not kwargs.get("filter") and not kwargs.get("input_csv") and cmd != "clear":
        # We're turning cache on/off for the whole datasource, so set it on the datasource item too
        _process_cache_command_on_item(cmd, datasource_item, client)

        print(
            "\nPersistent caching turned %s at datasource level. (All newly indexed items will have caching turned "
            "%s by default.)" % (cmd.upper(), cmd.upper())
        )


def _process_cache_command_on_item(cmd, item, client):
    items_api = sdk.ItemsApi(client)

    if cmd == "off":
        property_input = PropertyInputV1()
        property_input.value = False
        items_api.set_property(
            id=item.id, property_name="Cache Enabled", body=property_input
        )
    elif cmd == "on":
        property_input = PropertyInputV1()
        property_input.value = True
        items_api.set_property(
            id=item.id, property_name="Cache Enabled", body=property_input
        )
    elif cmd == "clear":
        items_api.clear_cache(id=item.id)


def map_same_server(**kwargs):
    """
    Create a map of items in a datasource to other items (probably in a different
    datasource).
    :param datasource_name: Datasource Name of the OLD datasource to search.
    :param datasource_class: Datasource Class of the OLD datasource to search.
    :param datasource_id: Datasource ID of the OLD datasource to search.
    :param id: Seeq ID of the OLD datasource to search.
    :param name_regex: Regular expression that matches on OLD item name.
    :param description_regex: Regular expression that matches on OLD item description.
    :param data_id_regex: Regular expression that matches on OLD Data ID.
    :param new_datasource_class: Datasource Class of datasource containing NEW item.
    :param new_datasource_id: Datasource ID of datasource containing NEW item.
    :param new_id: Seeq ID of datasource containing NEW item.
    :param new_name_regex: Regular expression that matches on NEW item name.
    :param new_description_regex: Regular expression that matches on NEW item description.
    :param new_data_id_regex: Regular expression that matches on NEW Data ID.
    :param csv: CSV file to write items to.
    :param append_csv: Append to the existing CSV file if possible.
    """

    client = spy.client
    items_api = sdk.ItemsApi(client)

    datasource_item = get_datasource_item(client, **kwargs)

    _filter = "Datasource Class==" + datasource_item.datasource_class
    _filter += " && Datasource ID==" + datasource_item.datasource_id

    if kwargs.get("name_regex"):
        _filter += " && Name ~= /%s/" % _remove_regex_capture_group_names(
            kwargs.get("name_regex")
        )

    if kwargs.get("description_regex"):
        _filter += " && Description ~= /%s/" % _remove_regex_capture_group_names(
            kwargs.get("description_regex")
        )

    if kwargs.get("data_id_regex"):
        _filter += " && Data ID ~= /%s/" % _remove_regex_capture_group_names(
            kwargs.get("data_id_regex")
        )

    new_datasource_args = {
        "datasource_class": kwargs.get("new_datasource_class"),
        "datasource_id": kwargs.get("new_datasource_id"),
        "id": kwargs.get("new_id"),
    }

    new_datasource_item = get_datasource_item(client, **new_datasource_args)

    tables.print_table_from_sdk_output(
        lambda offset: items_api.search_items(
            filters=[_filter, SeeqNames.API.Flags.include_unsearchable],
            offset=offset,
            limit=helper.get_default_page_size(),
            order_by=SORT_BY_ITEM_GUID,
        ),
        "items",
        [
            ColumnSpec("count", "#", ">", 8),
            ColumnSpec("type", "Type", "<", 15, console=False),
            ColumnSpec(
                "ancestors",
                "Old Path",
                ">",
                40,
                lambda ancestors, item: _ancestor_path(ancestors),
            ),
            ColumnSpec("name", "Old Name", "<", 40),
            ColumnSpec("description", "Old Description", "<", 40, console=False),
            ColumnSpec("value_unit_of_measure", "Old UOM", "<", 8, console=False),
            ColumnSpec("id", "Old ID", "<", 36, console=False),
            ColumnSpec("new_ancestors", "New Path", ">", 40),
            ColumnSpec("new_name", "New Name", "<", 40),
            ColumnSpec("new_description", "New Description", "<", 40, console=False),
            ColumnSpec("new_value_unit_of_measure", "New UOM", "<", 8, console=False),
            ColumnSpec("new_id", "New ID", "<", 36, console=False),
        ],
        lambda item: _process_map_same_server_command_on_item(item, client, new_datasource_item, **kwargs),
        csv_file_name=kwargs.get("csv"),
        append_csv=kwargs.get("append_csv"),
    )


def _remove_regex_capture_group_names(regex):
    # PostgreSQL will choke on named capture groups
    return re.sub(r"\?<\w+>", "", regex)


def _pythonize_regex_capture_group_names(regex):
    # Unlike standard regex syntax, Python capture groups have a capital P:  (?P<group_name>.*?)
    return re.sub(r"\?<(\w+)>", r"?P<\1>", regex)


def _process_map_same_server_command_on_item(item, client, new_datasource_item, **kwargs):
    items_api = sdk.ItemsApi(client)
    item_type = None

    new_filter = 'Datasource Class==' + new_datasource_item.datasource_class
    new_filter += ' && Datasource ID==' + new_datasource_item.datasource_id

    simple_match = False
    if (
            not kwargs.get("name_regex")
            and not kwargs.get("description_regex")
            and not kwargs.get("data_id_regex")
            and not kwargs.get("new_name_regex")
            and not kwargs.get("new_description_regex")
            and not kwargs.get("new_data_id_regex")
    ):
        # This is "identical datasources" mode, where the names of the items in each datasource are the same.
        # So we just filter on the name, assuming it's unique.
        new_filter += " && Name==%s" % item.name

        # Storing this and using it in the where clause instead of
        # using a filter results in a faster query that uses less RAM.
        item_type = item.type
        simple_match = True
    else:
        capture_groups = dict()
        new_name_regex = None

        def _update_capture_groups(_regex, _prop):
            _match = re.match(_pythonize_regex_capture_group_names(_regex), _prop)
            if _match:
                capture_groups.update(_match.groupdict())

        if kwargs.get("name_regex"):
            _update_capture_groups(kwargs.get("name_regex"), item.name)

        if kwargs.get("new_name_regex"):
            new_name_regex = kwargs.get("new_name_regex")
            new_name_regex = _replace_tokens_in_regex(new_name_regex, capture_groups)

        if new_name_regex:
            new_filter += " && Name ~= /%s/" % new_name_regex

        new_description_regex = None

        if kwargs.get("description_regex"):
            _update_capture_groups(kwargs.get("description_regex"), item.description)

        if kwargs.get("new_description_regex"):
            new_description_regex = kwargs.get("new_description_regex")
            new_description_regex = _replace_tokens_in_regex(
                new_description_regex, capture_groups
            )

        if new_description_regex:
            new_filter += " && Description ~= /%s/" % new_description_regex

        new_data_id_regex = None

        if kwargs.get("data_id_regex"):
            data_id_property = items_api.get_property(
                id=item.id, property_name=SeeqNames.Properties.data_id
            )  # type: PropertyOutputV1
            _update_capture_groups(kwargs.get("data_id_regex"), data_id_property.value)

        if kwargs.get("new_data_id_regex"):
            new_data_id_regex = kwargs.get("new_data_id_regex")
            new_data_id_regex = _replace_tokens_in_regex(
                new_data_id_regex, capture_groups
            )

        if new_data_id_regex:
            new_filter += " && Data ID == %s" % new_data_id_regex

    offset = 0
    matching_items = list()
    while True:
        search_results = items_api.search_items(
            filters=[new_filter, SeeqNames.API.Flags.include_unsearchable],
            offset=offset,
            limit=helper.get_default_page_size(),
            types=([item_type]) if item_type is not None else [],
        )  # type: ItemSearchPreviewPaginatedListV1

        for matching_item in search_results.items:
            if not simple_match or _ancestor_path(
                    matching_item.ancestors
            ) == _ancestor_path(item.ancestors):
                matching_items.append(matching_item)

        if len(search_results.items) < helper.get_default_page_size():
            break

        offset += len(search_results.items)

    if len(matching_items) == 1:
        new_item = matching_items[0]

        return [
            ("new_id", new_item.id),
            ("new_ancestors", _ancestor_path(new_item.ancestors)),
            ("new_name", new_item.name),
            ("new_description", new_item.description),
            ("new_value_unit_of_measure", new_item.value_unit_of_measure),
        ]

    elif len(matching_items) > 1:
        return [
            ("new_id", "MULTIPLE MATCHES"),
            ("new_ancestors", "MULTIPLE MATCHES"),
            ("new_name", "MULTIPLE MATCHES"),
            ("new_description", "MULTIPLE MATCHES"),
            ("new_value_unit_of_measure", ""),
        ]
    else:
        return [
            ("new_id", "NO MATCHES"),
            ("new_ancestors", "NO MATCHES"),
            ("new_name", "NO MATCHES"),
            ("new_description", "NO MATCHES"),
            ("new_value_unit_of_measure", ""),
        ]


def _replace_tokens_in_regex(regex, group_dict):
    for name, value in group_dict.items():
        regex = regex.replace("${%s}" % name, value)

    return regex


def map_old_server(**kwargs):
    """
    Create a list of items to be mapped later using "datasource_map_new_server()".
    :param datasource_name: Datasource Name of the OLD datasource to search.
    :param datasource_class: Datasource Class of the OLD datasource to search.
    :param datasource_id: Datasource ID of the OLD datasource to search.
    :param id: Seeq ID of the OLD datasource to search.
    :param name_regex: Regular expression that matches on OLD item name.
    :param description_regex: Regular expression that matches on OLD item description.
    :param data_id_regex: Regular expression that matches on OLD Data ID.
    :param csv: CSV file to write items to.
    :param append_csv: Append to the existing CSV file if possible.
    """
    client = spy.client

    items_api = sdk.ItemsApi(client)

    datasource_item = get_datasource_item(client, **kwargs)

    _filter = "Datasource Class==" + datasource_item.datasource_class
    _filter += " && Datasource ID==" + datasource_item.datasource_id

    if kwargs.get("name_regex"):
        _filter += " && Name ~= /%s/" % _remove_regex_capture_group_names(
            kwargs.get("name_regex")
        )

    if kwargs.get("description_regex"):
        _filter += " && Description ~= /%s/" % _remove_regex_capture_group_names(
            kwargs.get("description_regex")
        )

    if kwargs.get("data_id_regex"):
        _filter += " && Data ID ~= /%s/" % _remove_regex_capture_group_names(
            kwargs.get("data_id_regex")
        )

    tables.print_table_from_sdk_output(
        lambda offset: items_api.search_items(
            filters=[_filter, SeeqNames.API.Flags.include_unsearchable],
            offset=offset,
            limit=helper.get_default_page_size(),
            order_by=SORT_BY_ITEM_GUID,
        ),
        "items",
        [
            ColumnSpec("count", "#", ">", 8),
            ColumnSpec("type", "Type", "<", 15),
            ColumnSpec("name", "Old Name", "<", 40),
            ColumnSpec("description", "Old Description", "<", 40),
            ColumnSpec("id", "Old ID", "<", 36, console=False),
            ColumnSpec("datasource_id", "Old Datasource ID", "<", 36, console=False),
            ColumnSpec("data_id", "Old Data ID", "<", 36, console=False),
        ],
        lambda item: _process_map_old_server_command_on_item(item, client, **kwargs),
        csv_file_name=kwargs.get("csv"),
        append_csv=kwargs.get("append_csv"),
    )


def _process_map_old_server_command_on_item(item, client, **kwargs):
    items_api = sdk.ItemsApi(client)

    datasource_id = items_api.get_property(
        id=item.id, property_name=SeeqNames.Properties.datasource_id
    )  # type: PropertyOutputV1

    data_id = items_api.get_property(
        id=item.id, property_name=SeeqNames.Properties.data_id
    )  # type: PropertyOutputV1

    return [("datasource_id", datasource_id.value), ("data_id", data_id.value)]


def map_new_server(**kwargs):
    """
    Map the items from datasource_map_old_server() to items on a new server.
    :param old_server_csv: Filename of CSV created by datasource_map_old())server
    :param datasource_id: Datasource ID of the OLD datasource to search.
    :param name_regex: Regular expression that matches on OLD item name.
    :param description_regex: Regular expression that matches on OLD item description.
    :param data_id_regex: Regular expression that matches on OLD Data ID.
    :param new_datasource_class: New datasource class to map to.
    :param new_datasource_id: ID of datasource containing NEW item.
    :param new_name_regex: Regular expression that matches on NEW item name.
    :param new_description_regex: Regular expression that matches on NEW item description.
    :param new_data_id_regex: Regular expression that matches on NEW Data ID.
    :param csv: CSV file to write items to.
    :param append_csv: Append to the existing CSV file if possible.
    """
    if not kwargs.get("datasource_id") or not kwargs.get("old_server_csv"):
        print("Parameters datasource_id and old_server_id are required.")
        return

    (header, rows) = tables.read_table(kwargs.get("old_server_csv"))

    old_items = dict()
    old_items["id_index"] = tables.find_column_index(header, "Old ID")
    old_items["datasource_id_index"] = tables.find_column_index(
        header, "Old Datasource ID"
    )
    old_items["name_index"] = tables.find_column_index(header, "Old Name")
    old_items["type_index"] = tables.find_column_index(header, "Type")
    old_items["description_index"] = tables.find_column_index(header, "Old Description")
    old_items["uom_index"] = tables.find_column_index(header, "Old UOM")
    old_items["data_id_index"] = tables.find_column_index(header, "Old Data ID")

    column_specs = [
        ColumnSpec("count", "#", ">", 8),
        ColumnSpec("type", "Type", "<", 15),
        ColumnSpec("name", "Old Name", "<", 40),
        ColumnSpec("description", "Old Description", "<", 40, console=False),
        ColumnSpec("id", "Old ID", "<", 36, console=False),
        ColumnSpec("datasource_id", "Old Datasource ID", "<", 36, console=False),
        ColumnSpec("data_id", "Old Data ID", "<", 36, console=False),
        ColumnSpec("new_name", "New Name", "<", 40),
        ColumnSpec("new_description", "New Description", "<", 40, console=False),
        ColumnSpec("new_id", "New ID", "<", 36, console=False),
        ColumnSpec("new_datasource_id", "New Datasource ID", "<", 36, console=False),
        ColumnSpec("new_data_id", "New Data ID", "<", 36, console=False),
    ]

    table = None
    try:
        table = tables.start_table(
            column_specs, kwargs.get("csv"), kwargs.get("append_csv")
        )

        for row in rows:
            if row[old_items["datasource_id_index"]] == kwargs.get("datasource_id"):
                item = _process_map_new_server_command_on_item(row, old_items, **kwargs)
                tables.write_table_row(table, item)

    finally:
        if table:
            tables.finish_table(table)


def _process_map_new_server_command_on_item(row, old_items, **kwargs):
    client = spy.client

    items_api = sdk.ItemsApi(client)

    new_filter = "Datasource ID==%s" % (
        kwargs.get("new_datasource_id")
        if kwargs.get("new_datasource_id")
        else kwargs.get("datasource_id")
    )

    new_filter += " && Datasource Class==%s" % (
        kwargs.get("new_datasource_class")
        if kwargs.get("new_datasource_class")
        else kwargs.get("datasource_class")
    )
    types_filter = []

    if (
            not kwargs.get("name_regex")
            and not kwargs.get("description_regex")
            and not kwargs.get("data_id_regex")
            and not kwargs.get("new_name_regex")
            and not kwargs.get("new_description_regex")
            and not kwargs.get("new_data_id_regex")
    ):
        # This is "identical datasources" mode, where the names of the items in each datasource are the same.
        # So we just filter on the name and type, assuming that combination is unique
        new_filter += " && Name==%s" % row[old_items["name_index"]]
        types_filter = [row[old_items["type_index"]]]
    else:
        capture_groups = dict()
        new_name_regex = None

        if kwargs.get("name_regex"):
            capture_groups.update(
                re.match(
                    _pythonize_regex_capture_group_names(kwargs.get("name_regex")),
                    row[old_items["name_index"]],
                ).groupdict()
            )

        if kwargs.get("new_name_regex"):
            new_name_regex = kwargs.get("new_name_regex")
            new_name_regex = _replace_tokens_in_regex(new_name_regex, capture_groups)

        if new_name_regex:
            new_filter += " && Name ~= /%s/" % new_name_regex

        new_description_regex = None

        if kwargs.get("description_regex"):
            capture_groups.update(
                re.match(
                    _pythonize_regex_capture_group_names(
                        kwargs.get("description_regex")
                    ),
                    row[old_items["description_index"]],
                ).groupdict()
            )

        if kwargs.get("new_description_regex"):
            new_description_regex = kwargs.get("new_description_regex")
            new_description_regex = _replace_tokens_in_regex(
                new_description_regex, capture_groups
            )

        if new_description_regex:
            new_filter += " && Description ~= /%s/" % new_description_regex

        new_data_id_regex = None

        if kwargs.get("data_id_regex"):
            capture_groups.update(
                re.match(
                    _pythonize_regex_capture_group_names(kwargs.get("data_id_regex")),
                    row[old_items["data_id_index"]],
                ).groupdict()
            )

        if kwargs.get("new_data_id_regex"):
            new_data_id_regex = kwargs.get("new_data_id_regex")
            new_data_id_regex = _replace_tokens_in_regex(
                new_data_id_regex, capture_groups
            )

        if new_data_id_regex:
            new_filter += " && Data ID == %s" % new_data_id_regex

    search_results = items_api.search_items(
        filters=[new_filter, SeeqNames.API.Flags.include_unsearchable],
        types=types_filter,
        offset=0,
        limit=2,
    )  # type: ItemSearchPreviewPaginatedListV1

    if len(search_results.items) == 1:
        new_item = search_results.items[0]

        datasource_id = items_api.get_property(
            id=new_item.id, property_name=SeeqNames.Properties.datasource_id
        )  # type: PropertyOutputV1

        data_id = items_api.get_property(
            id=new_item.id, property_name=SeeqNames.Properties.data_id
        )  # type: PropertyOutputV1

        new_id = new_item.id
        new_name = new_item.name
        new_type = new_item.type
        new_description = new_item.description
        new_datasource_id = datasource_id.value
        new_data_id = data_id.value
    elif len(search_results.items) > 1:
        new_id = "MULTIPLE MATCHES"
        new_name = "MULTIPLE MATCHES"
        new_type = "MULTIPLE MATCHES"
        new_description = "MULTIPLE MATCHES"
        new_datasource_id = "MULTIPLE MATCHES"
        new_data_id = "MULTIPLE MATCHES"
    else:
        new_id = "NO MATCHES"
        new_name = "NO MATCHES"
        new_type = "NO MATCHES"
        new_description = "NO MATCHES"
        new_datasource_id = "NO MATCHES"
        new_data_id = "NO MATCHES"

    return shared.Bunch(
        id=row[old_items["id_index"]],
        name=row[old_items["name_index"]],
        type=row[old_items["type_index"]],
        description=row[old_items["description_index"]],
        datasource_id=row[old_items["datasource_id_index"]],
        data_id=row[old_items["data_id_index"]],
        new_id=new_id,
        new_type=new_type,
        new_name=new_name,
        new_description=new_description,
        new_datasource_id=new_datasource_id,
        new_data_id=new_data_id,
    )


def _process_cache_fill_on_item(item, client, start, end, ignore_errors):
    formulas_api = sdk.FormulasApi(client)

    formula = (
            'group(capsule("%s","%s")).toTable("stats").addStatColumn("series",$series,count())'
            % (start, end)
    )

    parameter = "series=%s" % item.id

    while True:
        try:
            formula_run_output, _, http_headers = (
                formulas_api.run_formula_with_http_info(
                    formula=formula, parameters=[parameter]
                )
            )
            break
        except MaxRetryError:
            # Server is down, keep trying in this loop
            print("Seeq Server appears to be down, trying to reconnect")
        except ApiException as e:
            if e.status == 401:
                # This can happen if the server is restarted while the script is running
                print("Logging in again")
                client = spy.client
            else:
                print("Seeq Server error encountered: %s" % e)

        except Exception as e:
            print("Seeq Server error encountered: %s" % e)

        if ignore_errors:
            # If we're asked to ignore errors, then stop retrying the current item and just log an error to move on.
            # This prevents the script from getting hung up on a single item.
            return [("new_samples", "ERROR"), ("all_samples", "ERROR")]

        time.sleep(5)

    all_samples_in_range = int(formula_run_output.table.data[0][2])

    server_meters_string = http_headers[SeeqNames.API.Headers.server_meters]

    samples_read_from_datasource = 0
    server_meters = server_meters_string.split(",")
    for server_meter_string in server_meters:
        server_meter = server_meter_string.split(";")
        if len(server_meter) < 3:
            continue

        dur_string = server_meter[1]
        desc_string = server_meter[2]

        dur_value = int(dur_string.split("=")[1])

        if (
                desc_string
                == f'desc="{SeeqNames.API.Headers.Meters.datasource_samples_read}"'
        ):
            samples_read_from_datasource = int(dur_value)
            break

    return [
        ("new_samples", samples_read_from_datasource),
        # New samples may be less than all samples due to boundary values, so take the max below so that it isn't
        # 'incongruent' in users' eyes.
        ("all_samples", max(all_samples_in_range, samples_read_from_datasource)),
    ]


def _get_cache_fill_command_item_filter(client, **kwargs):
    datasource_item = get_datasource_item(client, **kwargs)

    _filter = " && Datasource Class==" + datasource_item.datasource_class
    _filter += " && Datasource ID==" + datasource_item.datasource_id

    _filter += " && Cache Enabled == true"

    return _filter


def _cache_file_input_csv(**kwargs):
    try:
        if not util.safe_exists(kwargs.get("input_csv")):
            print('Path "%s" does not exist.' % kwargs.get("input_csv"))
            return list(), 0

        if util.safe_isdir(kwargs.get("input_csv")):
            input_files = [
                os.path.join(kwargs.get("input_csv"), f)
                for f in os.listdir(kwargs.get("input_csv"))
            ]
        else:
            input_files = [kwargs.get("input_csv")]

        latest_mtime = 0
        for input_file in input_files:
            latest_mtime = max(latest_mtime, os.path.getmtime(input_file))

        return input_files, latest_mtime

    except Exception as e:
        print("Exception encountered: %s" % e)
        return list(), 0


def _cache_fill_items_from_item_id(client, item_id, _types, item_name):
    items_api = sdk.ItemsApi(client)

    try:
        item = items_api.get_item_and_all_properties(id=item_id)  # type: ItemOutputV1
        if item.type not in _types:
            return list()

        cache_enabled = False
        for _property in item.properties:  # type: PropertyOutputV1
            if _property.name == "Cache Enabled" and _property.value == "true":
                cache_enabled = True
                break

        if not cache_enabled:
            return list()

        return [
            shared.Bunch(
                name=item.name,
                type=item.type,
                id=item.id,
                new_samples=0,
                all_samples=0,
                total_time_str="",
                action_time_str="",
                start="",
                end="",
            )
        ]
    except Exception:
        return [
            shared.Bunch(
                name=item_name,
                type="",
                id="NOT FOUND",
                new_samples=0,
                all_samples=0,
                total_time_str="",
                action_time_str="",
                start="",
                end="",
            )
        ]


def _cache_fill_items_from_item_name(client, _filter, _types, item_name, **kwargs):
    items_api = sdk.ItemsApi(client)

    if not _filter:
        _filter = _get_cache_fill_command_item_filter(client, **kwargs)

    search_results = items_api.search_items(
        filters=[_filter + "&& Name == " + item_name], types=_types
    )  # type: ItemSearchPreviewPaginatedListV1
    if len(search_results.items) > 0:
        return search_results.items
    else:
        return [
            shared.Bunch(
                name=item_name,
                type="",
                id="NOT FOUND",
                new_samples=0,
                all_samples=0,
                total_time_str="",
                action_time_str="",
                start="",
                end="",
            )
        ]


class InputFilesChanged(Exception):
    def __init__(self):
        pass


def _cache_fill_process_via_input_files(
        client, _types, _filter, current_start_iso, current_end_iso, **kwargs
):
    input_files, latest_mtime = _cache_file_input_csv(**kwargs)

    if len(input_files) == 0:
        print("No CSV files to process.")
        time.sleep(5)
        return

    column_specs = [
        ColumnSpec("count", "#", ">", 8),
        ColumnSpec("action_time_str", "Time", ">", 8),
        ColumnSpec("total_time_str", "Total", ">", 8),
        ColumnSpec("new_samples", "New", ">", 8),
        ColumnSpec("all_samples", "All", ">", 8),
        ColumnSpec("start", "Start", "<", 20),
        ColumnSpec("end", "End", "<", 20),
        ColumnSpec("name", "Name", "<", 40),
        ColumnSpec("type", "Type", "<", 15, console=False),
        ColumnSpec("id", "ID", "<", 36),
    ]

    try:
        for input_file in input_files:
            print('Processing "%s"...' % input_file)

            total_start = time.time()

            header, rows = tables.read_table(input_file)

            id_column = tables.find_column_index(header, "ID")
            name_column = tables.find_column_index(header, "Name")
            duration_column = tables.find_column_index(header, "Duration")

            if name_column is None and id_column is None:
                raise RuntimeError('"ID" nor "Name" column not found in input CSV file')

            table = tables.start_table(column_specs, kwargs.get("csv"))

            try:
                for row in rows:
                    try:
                        new_input_files, new_latest_mtime = _cache_file_input_csv(
                            **kwargs
                        )
                        if kwargs.get("continuous") and (
                                latest_mtime != new_latest_mtime
                                or input_files != new_input_files
                        ):
                            raise InputFilesChanged()

                        item_id = row[id_column] if id_column is not None else None
                        item_name = (
                            row[name_column] if name_column is not None else None
                        )

                        if duration_column is not None:
                            duration = int(row[duration_column])
                            this_end = datetime.now(utc)
                            this_start = this_end - timedelta(hours=duration)
                            current_start_iso = this_start.strftime(
                                "%Y-%m-%dT%H:%M:%SZ"
                            )
                            current_end_iso = this_end.strftime("%Y-%m-%dT%H:%M:%SZ")

                        items = []
                        if item_id:
                            items = _cache_fill_items_from_item_id(
                                client, item_id, _types, item_name
                            )
                        elif item_name:
                            items = _cache_fill_items_from_item_name(
                                client, _filter, _types, item_name, **kwargs
                            )

                        for _item in items:
                            action_start = time.time()
                            if _item.id != "NOT FOUND":
                                fill_metrics = _process_cache_fill_on_item(
                                    _item,
                                    client,
                                    current_start_iso,
                                    current_end_iso,
                                    kwargs.get("ignore_errors"),
                                )
                                action_time_sec = time.time() - action_start
                                total_time_sec = time.time() - total_start
                                _, new_samples = fill_metrics[0]
                                _, all_samples = fill_metrics[1]
                                _item = shared.Bunch(
                                    name=_item.name,
                                    type=_item.type,
                                    id=_item.id,
                                    new_samples=str(new_samples),
                                    all_samples=str(all_samples),
                                    total_time_str=tables._print_time_string(
                                        total_time_sec
                                    ),
                                    action_time_str=tables._print_time_string(
                                        action_time_sec
                                    ),
                                    start=current_start_iso,
                                    end=current_end_iso,
                                )
                            tables.write_table_row(table, _item)
                    except InputFilesChanged:
                        raise
                    except Exception as e:
                        if kwargs.get("continuous") and kwargs.get("ignore_errors"):
                            print(
                                "Error encountered on row, ignoring:\n%s\n%s" % (row, e)
                            )
                        else:
                            raise
            finally:
                tables.finish_table(table)

    except InputFilesChanged:
        print('File(s) "%s" have changed, restarting' % kwargs.get("input_csv"))


def _cache_fill_process_via_item_search(
        client, _types, _filter, current_start_iso, current_end_iso, **kwargs
):
    items_api = sdk.ItemsApi(client)

    print(
        "\nLooping through items to cache from %s to %s\n"
        % (current_start_iso, current_end_iso)
    )

    column_specs = [
        ColumnSpec("count", "#", ">", 8),
        ColumnSpec("action_time_str", "Time", ">", 8),
        ColumnSpec("total_time_str", "Total", ">", 8),
        ColumnSpec("new_samples", "New", ">", 8),
        ColumnSpec("all_samples", "All", ">", 8),
        ColumnSpec(
            "ancestors",
            "Path",
            ">",
            40,
            lambda ancestors, _item: _ancestor_path(ancestors),
        ),
        ColumnSpec("name", "Name", "<", 40),
        ColumnSpec("type", "Type", "<", 15, console=False),
        ColumnSpec("id", "ID", "<", 36, console=False),
    ]

    if not _filter:
        _filter = _get_cache_fill_command_item_filter(client, **kwargs)

    if kwargs.get("filter"):
        _filter += " && Name ~= " + kwargs.get("filter")

    tables.print_table_from_sdk_output(
        lambda offset: items_api.search_items(
            filters=[_filter],
            offset=offset,
            types=_types,
            limit=helper.get_default_page_size(),
            order_by=SORT_BY_ITEM_GUID,
        ),
        "items",
        column_specs,
        lambda _item: _process_cache_fill_on_item(
            _item,
            client,
            current_start_iso,
            current_end_iso,
            kwargs.get("ignore_errors"),
        ),
        csv_file_name=kwargs.get("csv"),
    )


def cache_fill(**kwargs):
    """
    Fill the cache with data from the datasource.
    :param datasource_name: Datasource Name of datasource to search.
    :param datasource_class: Datasource Class of datasource to search.
    :param datasource_id: Datasource ID of datasource to search.
    :param id: Seeq ID of datasource to search.
    :param input_csv: CSV file with ID or Name column for items to affect.
    :param start: start date for fill (e.g. 2018-08-02T00:00:00Z).
    :param end: end date for fill (e.g. 2018-09-31T00:00:00Z).
    :param continuous: continuously fill the cache (Do not specify end).
    :param ignore_errors: ignore errors while accessing data.
    :param minimum: minimum query size (in hours).
    :param csv: name of CSV file to route output to.
    """
    kwargs["continuous"] = kwargs.get("continuous") or False
    kwargs["ignore_errors"] = kwargs.get("ignore_errors") or False

    client = spy.client

    current_start = (
        dateutil.parser.parse(kwargs.get("start"))
        if kwargs.get("start")
        else datetime.now(utc)
    )

    _types = ["StoredSignal", "CalculatedSignal"]
    _filter = None

    minimum = int(kwargs.get("minimum") or 0)

    while True:
        current_end = (
            dateutil.parser.parse(kwargs.get("end"))
            if kwargs.get("end")
            else datetime.now(utc)
        )

        try:
            time_delta = current_end - current_start
            if (time_delta.total_seconds() / 60 / 60) < minimum:
                current_start = current_end - timedelta(hours=minimum)

            current_start_iso = current_start.strftime("%Y-%m-%dT%H:%M:%SZ")
            current_end_iso = current_end.strftime("%Y-%m-%dT%H:%M:%SZ")

            if kwargs.get("filter") and kwargs.get("input_csv"):
                raise RuntimeError(
                    "filter and input-csv arguments are mutually exclusive"
                )

            if kwargs.get("input_csv"):
                _cache_fill_process_via_input_files(
                    client,
                    _types,
                    _filter,
                    current_start_iso,
                    current_end_iso,
                    **kwargs,
                )
            else:
                _cache_fill_process_via_item_search(
                    client,
                    _types,
                    _filter,
                    current_start_iso,
                    current_end_iso,
                    **kwargs,
                )

        except ApiException as e:
            if e.status == 401:
                print("Logging in again")
                client = spy.client
            else:
                if kwargs.get("continuous") and kwargs.get("ignore_errors"):
                    print("Error encountered, ignoring: %s" % e)
                else:
                    raise
        except Exception as e:
            if kwargs.get("continuous") and kwargs.get("ignore_errors"):
                print("Error encountered, ignoring: %s" % e)
            else:
                raise

        if not kwargs.get("continuous"):
            break

        current_start = current_end

        time.sleep(5)

    if not kwargs.get("continuous"):
        print("\nPersistent cache FILLED for all of the above items.")
        print("Note: Only items with caching ON were affected.")
        print(
            'Note: Time ranges near "now" may not be fully cached to allow for soon-arriving data.'
        )



from __future__ import annotations

import numpy as np
import pandas as pd

from seeq.spy import _common
from seeq.spy._errors import SPyDependencyNotFound


def plot(samples, *, capsules=None, size=None, show=True):
    """
    Plots signals/samples via matplotlib, optionally including conditions/
    capsules as shaded regions.

    Parameters
    ----------
    samples : pd.DataFrame
        A DataFrame with a pandas.Timestamp-based index with a column for
        each signal.

    capsules : pd.DataFrame, optional
        A DataFrame with (at minimum) three columns: Condition, Capsule Start
        and Capsule End. Each unique Condition will be plotted as a
        differently-colored shaded region behind the sample data.

    size : str, optional
        This value, if provided, is passed directly to
        matplotlib.rcParams['figure.figsize'] to control the size of the
        rendered plot.

    show : bool, default True
        Set this to False if you don't actually want to show the plot. Used
        mainly for testing.
    """
    _common.validate_argument_types([
        (samples, 'samples', pd.DataFrame),
        (capsules, 'capsules', pd.DataFrame),
        (size, 'size', str)
    ])

    if size:
        try:
            import matplotlib
        except ImportError:
            raise SPyDependencyNotFound(
                f'`matplotlib` is required to use this feature. Please use '
                f'`pip install seeq-spy[templates]` to use this feature.')
        matplotlib.rcParams['figure.figsize'] = size

    def _convert_to_timestamp(matplotlib_timestamp, matplotlib_axis):
        return pd.Period(ordinal=int(matplotlib_timestamp), freq=matplotlib_axis.freq).to_timestamp()

    # matplotlib basically has difficulty with timezones. So it's better to make the index timezone-naive and take it
    # out of matplotlib's hands.
    timezone = samples.index.tz
    samples.index = samples.index.tz_localize(None)

    ax = samples.plot()

    if capsules is not None:
        try:
            from matplotlib.pyplot import cm
        except ImportError:
            raise SPyDependencyNotFound(
                f'`matplotlib` is required to use this feature. Please use '
                f'`pip install seeq-spy[templates]` to use this feature.')
        unique_conditions = capsules[['Condition']].drop_duplicates()['Condition'].to_list()

        capsule_colors = dict()
        colors = cm.tab10(np.linspace(0, 1, len(unique_conditions)))
        for i in range(0, len(colors)):
            capsule_colors[unique_conditions[i]] = colors[i]

        axis_start_matplotlib, axis_end_matplotlib = ax.get_xlim()
        axis_start = _convert_to_timestamp(axis_start_matplotlib, ax)
        axis_end = _convert_to_timestamp(axis_end_matplotlib, ax)

        for index, capsule in capsules.iterrows():
            color = capsule_colors[capsule['Condition']]
            start = axis_start if pd.isna(capsule['Capsule Start']) else capsule['Capsule Start']
            if start.tz:
                start = start.tz_convert(timezone).tz_localize(None)
            end = axis_end if pd.isna(capsule['Capsule End']) else capsule['Capsule End']
            if end.tz:
                end = end.tz_convert(timezone).tz_localize(None)

            ax.axvspan(start, end, facecolor=color, edgecolor=None, alpha=0.3)

    if show:
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise SPyDependencyNotFound(
                f'`matplotlib` is required to use this feature. Please use '
                f'`pip install seeq-spy[templates]` to use this feature.')
        plt.show()



from __future__ import annotations

import os
import re
import types
from typing import Dict, Optional, Union

import pandas as pd

from seeq import spy
from seeq.base import util
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._session import Session
from seeq.spy._status import Status


@Status.handle_keyboard_interrupt()
def pull(job_folder: str, workbooks_df: Union[pd.DataFrame, str], *, resume: bool = True,
         include_referenced_workbooks: bool = True,
         include_rendered_content: bool = False,
         errors: Optional[str] = None, quiet: Optional[bool] = None,
         status: Optional[Status] = None,
         session: Optional[Session] = None) -> pd.DataFrame:
    """
    Pulls the definitions for each workbook specified by workbooks_df on to
    disk, in a restartable "job"-like fashion.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to a folder on disk where the workbooks
        definitions will be saved. If the folder does not exist, it will be
        created. If the folder exists, the job will continue where it left off.

    workbooks_df : {str, pd.DataFrame}
        A DataFrame containing 'ID', 'Type' and 'Workbook Type' columns that
        can be used to identify the workbooks to pull. This is usually created
        via a call to spy.workbooks.search(). Alternatively, you can supply a
        workbook ID directly as a str or the URL of a Seeq Workbench worksheet
        as a str.

    resume : bool, default True
        True if the pull should resume from where it left off, False if it
        should pull everything again.

    include_referenced_workbooks : bool, default True
        If True, Analyses that are depended upon by Topics will be
        automatically included in the resulting list even if they were not part
        of the workbooks_df DataFrame.

    include_rendered_content : bool, default False
        If True, any Organizer Topics pulled will include rendered content
        images, which will cause spy.workbooks.save() to include a folder for
        each Topic Document with its HTML and images such that it can be loaded
        and viewed in a browser.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame (errors='catalog' must be combined with
        status=<Status object>).

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    """
    input_args = _common.validate_argument_types([
        (job_folder, 'job_folder', str),
        (workbooks_df, 'workbooks_df', (pd.DataFrame, str)),
        (resume, 'resume', bool),
        (include_referenced_workbooks, 'include_referenced_workbooks', bool),
        (include_rendered_content, 'include_rendered_content', bool),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    _login.validate_login(session, status)

    util.safe_makedirs(job_folder, exist_ok=True)

    job_workbooks_folder = get_workbooks_folder(job_folder)
    util.safe_makedirs(job_workbooks_folder, exist_ok=True)

    existing = {workbook_id: workbook_folder for workbook_folder, workbook_id in
                walk_workbook_folders(job_workbooks_folder)}

    # Remove existing workbook folders that did not complete successfully
    to_remove = set()
    for workbook_id in existing:
        workbook_folder = os.path.join(job_workbooks_folder, existing[workbook_id])
        if not util.safe_exists(os.path.join(workbook_folder, 'Complete')):
            # Something must have gone wrong during the save. Delete the whole thing so we start over
            to_remove.add(workbook_id)

    for workbook_id in to_remove:
        util.safe_rmtree(os.path.join(job_workbooks_folder, existing[workbook_id]))
        del existing[workbook_id]

    def _load_if_exists(_workbook_id: str):
        if resume and _workbook_id in existing:
            return spy.workbooks.load(os.path.join(job_workbooks_folder, existing[_workbook_id]), quiet=True)[0]
        else:
            return None

    data_usages: Dict[str, dict] = dict()
    all_workstep_usages: Dict[str, Dict[str, Union[list, set]]] = dict()

    def _ensure(_workstep_id):
        if _workstep_id not in all_workstep_usages:
            all_workstep_usages[_workstep_id] = {'Periods': list(), 'Stored Items': set()}

    for workbook in spy.workbooks._pull.do_pull(
            workbooks_df, status=status, session=session,
            include_referenced_workbooks=include_referenced_workbooks,
            include_rendered_content=include_rendered_content,
            preprocess_callback=_load_if_exists):
        workstep_usages = workbook.get_workstep_usages(now=pd.Timestamp.utcnow())
        for workstep_id, periods in workstep_usages.items():
            _ensure(workstep_id)
            all_workstep_usages[workstep_id]['Periods'].extend(periods)

        if isinstance(workbook, spy.workbooks.Analysis):
            for worksheet in workbook.worksheets:
                for workstep_id, workstep in worksheet.worksteps.items():
                    _ensure(workstep_id)
                    stored_item_references = workstep.get_stored_item_references(workbook.item_inventory)
                    all_workstep_usages[workstep_id]['Stored Items'].update({r.id for r in stored_item_references})
                    for r in stored_item_references:
                        data_usages[r.id] = {'Definition': r.definition_dict, 'Periods': list()}

        if not resume or workbook.id not in existing:
            spy.workbooks.save(workbook, job_workbooks_folder, overwrite=True, quiet=True)

    for workstep_id, workstep_usages in all_workstep_usages.items():
        for stored_item_id in workstep_usages['Stored Items']:
            if stored_item_id not in data_usages:
                raise Exception(f'Something went wrong-- no data usage found for workstep {stored_item_id}')
            data_usages[stored_item_id]['Periods'].extend(workstep_usages['Periods'])

    for entry in data_usages.values():
        entry['Periods'] = flatten_timestamps(entry['Periods'])

    spy.workbooks.job.data._pull.save_data_usage(job_folder, data_usages)

    copy_datasource_maps_to_root(job_folder)

    results_df = status.df.copy()

    results_df_properties = types.SimpleNamespace(
        func='spy.workbooks.job.pull',
        kwargs=input_args,
        status=status)

    _common.put_properties_on_df(results_df, results_df_properties)

    return results_df


def flatten_timestamps(timestamps):
    if not timestamps:
        return list()

    # Sort the timestamp pairs based on their start times
    sorted_timestamps = sorted(timestamps, key=lambda x: x['Start'])

    # Initialize the result list with the first timestamp pair
    flattened = [sorted_timestamps[0]]

    for current in sorted_timestamps[1:]:
        # Get the last merged interval from the result list
        last = flattened[-1]

        # Check for overlap and merge if needed
        if current['Start'] <= last['End']:
            # Merge the intervals by updating the end time
            flattened[-1] = {'Start': last['Start'], 'End': max(last['End'], current['End'])}
        else:
            # No overlap, add the current interval to the result list
            flattened.append({'Start': current['Start'], 'End': current['End']})

    return flattened


def copy_datasource_maps_to_root(job_folder):
    job_datasource_maps_folder = get_datasource_maps_folder(job_folder)
    job_workbooks_folder = get_workbooks_folder(job_folder)
    util.safe_makedirs(job_datasource_maps_folder, exist_ok=True)
    for workbook_folder, _, in walk_workbook_folders(job_workbooks_folder):
        workbook_folder = os.path.join(job_workbooks_folder, workbook_folder)
        for datasource_map_file in os.listdir(workbook_folder):  # type: str
            if not datasource_map_file.startswith('Datasource_Map_'):
                continue

            src = os.path.join(workbook_folder, datasource_map_file)
            dest = os.path.join(job_datasource_maps_folder, datasource_map_file)
            if not util.safe_exists(dest):
                util.safe_copy(src, dest)


def get_datasource_maps_folder(job_folder):
    job_datasource_maps_folder = os.path.join(job_folder, 'Datasource Maps')
    return job_datasource_maps_folder


def get_workbooks_folder(job_folder):
    return os.path.join(job_folder, 'Workbooks')


def walk_workbook_folders(job_workbooks_folder):
    for folder in os.listdir(job_workbooks_folder):
        path = os.path.join(job_workbooks_folder, folder)
        if not util.safe_isdir(path):
            continue

        match = re.match(r'^.*? \(([^)]+)\)$', folder)
        if match is None:
            continue

        yield match.group(0), match.group(1)


def redo(job_folder: str, status: Status):
    job_workbooks_folder = get_workbooks_folder(job_folder)
    workbook_folders = {i: f for f, i in walk_workbook_folders(job_workbooks_folder)}
    workbook_ids: pd.Series = status.df['ID']
    for index, workbook_id in workbook_ids.items():
        if workbook_id in workbook_folders:
            util.safe_rmtree(os.path.join(job_workbooks_folder, workbook_folders[workbook_id]))
            status.df.at[index, 'Result'] = 'Pull will be redone'
        else:
            status.df.at[index, 'Result'] = 'Not found'



from __future__ import annotations

import types
from datetime import datetime, timedelta, timezone
from typing import Optional

import pandas as pd

from seeq.sdk import *
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._errors import *
from seeq.spy._redaction import safely
from seeq.spy._session import Session
from seeq.spy._status import Status


@Status.handle_keyboard_interrupt()
def archive(items: pd.DataFrame,
            *,
            note: Optional[str] = None,
            undo: bool = False,
            errors: Optional[str] = None,
            quiet: Optional[bool] = None,
            status: Optional[Status] = None,
            session: Optional[Session] = None) -> pd.DataFrame:
    """
    Archives (moves to the trash) the items whose IDs are in the input
    DataFrame.

    Parameters
    ----------
    items : pandas.DataFrame
        A DataFrame representing the items that should be archived. Each row
        must include a valid GUID in the "ID" column.

    note : str, default 'Archived from SPy by <username> at <timestamp>.'
        The note that will be added to the archival event for historic
        observability.

    undo : bool, default False
        If True, the items will be unarchived (moved out of the trash) instead
        of archived. The 'note' argument will be ignored.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame.

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the metadata for the items that have been archived,
        along with any errors and statistics about the operation.

        Additionally, the following properties are stored on the "spy"
        attribute of the output DataFrame:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        func                A str value of 'spy.archive'
        kwargs              A dict with the values of the input parameters
                            passed to spy.archive to get the output DataFrame
        status              A spy.Status object with the status of the
                            spy.archive call
        =================== ===================================================

    Examples
    --------
    Search for signals with the name matching 'Well Head 00*' in the 'Custom Views' datasource and move
    them to the trash while ignoring any errors:

    >>> search_results = spy.search({'Name': 'Well Head 00*', 'Datasource Name': 'Custom Views'})
    >>> archive_results = spy.archive(search_results, errors='catalog')

    Pull the global asset tree with the name 'BioWorks Operations' and archive the entire heirarchy:

    >>> tree = spy.assets.Tree('BioWorks Operations', workbook=None)
    >>> archive_results = spy.archive(tree.df)

    Find all my workbooks with the name 'Condenser Maintenance' and trash them:

    >>> workbooks = spy.workbooks.search({'Name': 'Condenser Maintenance'}, content_filter='owner')
    >>> archive_results = spy.archive(workbooks)
    """
    input_args = _common.validate_argument_types([
        (items, 'items', pd.DataFrame),
        (note, 'note', str),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    _login.validate_login(session, status)
    activity = 'archiv' if not undo else 'unarchiv'

    if not _common.present(items, 'ID'):
        raise SPyValueError('The items DataFrame must contain an "ID" column.')

    if note is None:
        note = f'Archived from SPy by {session.user.username} at {datetime.now(timezone.utc).isoformat()}.'

    item_count = len(items)
    results_list = list()
    status.df = pd.DataFrame([{'Count': 0, 'Time': timedelta(0)}])
    status.update(f'Archiving 0 of {item_count} items', Status.RUNNING)

    items_api = ItemsApi(session.client)

    def _handle_row(_row: pd.Series):
        """
        Safely archives the item with the given ID.
        Adds the row with the additional 'Result' column to the results_list.
        """
        _item_id = _row['ID']

        def _add_error_message_and_warn(msg):
            _row['Result'] = msg
            results_list.append(_row)
            status.warn(msg)

        def _archive_row():
            if not _common.is_guid(_item_id):
                raise SPyValueError(f'ID "{_item_id}" is not a valid GUID')
            items_api.archive_item(id=_item_id, note=note)
            _row['Result'] = 'Success'
            results_list.append(_row)

        def _unarchive_row():
            if not _common.is_guid(_item_id):
                raise SPyValueError(f'ID "{_item_id}" is not a valid GUID')
            items_api.set_property(id=_item_id, property_name='Archived', body=PropertyInputV1(value=False))
            _row['Result'] = 'Success'
            results_list.append(_row)

        _func = _unarchive_row if undo else _archive_row
        safely(lambda: _func(),
               action_description=f'{activity}e item with ID {_item_id}',
               additional_errors=[400, 409],
               on_error=_add_error_message_and_warn,
               status=status)

    def _update_status(_index, _result):
        completion_count = len(results_list)
        if completion_count % 100 == 0:
            status.df.at[0, 'Count'] = completion_count
            status.df.at[0, 'Time'] = status.get_timer()
            status.update(f'{activity.capitalize()}ing {completion_count} of {item_count} items',
                          Status.RUNNING)

    for index, row in items.iterrows():
        status.add_job(index, (_handle_row, row), _update_status)

    status.execute_jobs(session, simple=True)

    archive_result_df = pd.DataFrame(results_list)

    status.df.at[0, 'Count'] = len(results_list)
    status.df.at[0, 'Time'] = status.get_timer()
    if 'Result' in archive_result_df.columns:
        success_count = len(archive_result_df[archive_result_df['Result'] == 'Success'])
    else:
        success_count = 0

    status.update(f'Successfully {activity}ed {success_count} of {item_count} items', Status.SUCCESS)
    archive_df_properties = types.SimpleNamespace(func='spy.archive', kwargs=input_args, status=status)
    _common.put_properties_on_df(archive_result_df, archive_df_properties)

    return archive_result_df



"""
seeq_onnx_helper.py

This module provides client and utilities for interacting with the Seeq Server and handling ONNX models.
It includes functionality for authentication, ONNX model description and model listing, registration, update and
archiving in Seeq.
"""

import base64
import io
import json
from abc import ABC
from typing import List, Dict
from urllib.parse import urljoin

import onnx
import pandas as pd
import requests


class PropertyNames:
    Name = "Name"
    Description = "Description"
    Inputs = "Inputs"
    ResultType = "Result Type"
    ScopedTo = "Scoped To"


class FormulaType:
    OnnxPredictionModel = "OnnxPredictionModel"
    OnnxAnomalyModel = "OnnxAnomalyModel"


def get_auth_token(host: str, username: str, password: str) -> str:
    """
    Authenticates with the Seeq server and retrieves an authentication token.

    Parameters:
        host (str): The base URL of the Seeq server.
        username (str): The username for authentication.
        password (str): The password for authentication.

    Returns:
        str: The authentication token (x-sq-auth).

    Example:
        >>> token = get_auth_token("https://tenant.seeq.com", "my_username", "my_password")
    """
    endpoint = urljoin(host, "/api/auth/login")
    response = requests.post(
        endpoint,
        headers={
            "Content-Type": "application/vnd.seeq.v1+json",
            "Accept": "application/vnd.seeq.v1+json",
        },
        data=json.dumps(
            {
                "username": username,
                "password": password,
                "authProviderClass": "Auth",
                "authProviderId": "Seeq",
            }
        ),
    )
    response.raise_for_status()
    return response.headers["x-sq-auth"]


def describe_onnx(model: onnx.ModelProto) -> None:
    """
    Prints a description of the ONNX model, including its inputs and outputs.

    Parameters:
        model (onnx.ModelProto): The ONNX model to describe.

    Returns:
        None

    Example:
        >>> import onnx
        >>> model = onnx.load("model.onnx")
        >>> describe_onnx(model)
    """

    def repr_tensor(tensor):
        name = tensor.name
        elem_type = onnx.TensorProto.DataType.Name(tensor.type.tensor_type.elem_type)
        shape = [
            dim.dim_value if (dim.dim_value > 0) else "?"
            for dim in tensor.type.tensor_type.shape.dim
        ]
        print(f"  - {name}: type={elem_type}, shape={shape}")

    graph = model.graph
    print(f"\n=== Model: {graph.name} ===")

    # Inputs
    print("\nInputs:")
    for input_tensor in graph.input:
        repr_tensor(input_tensor)

    # Outputs
    print("\nOutputs:")
    for output_tensor in graph.output:
        repr_tensor(output_tensor)


def cast_double_to_float(model: onnx.ModelProto) -> onnx.ModelProto:
    """
    Modifies an ONNX model to cast its input from double (float64) to float (float32).

    This function inserts a `Cast` node at the beginning of the model's graph to convert
    the input tensor's data type from double (float64) to float (float32). It also updates
    the input tensor's type and adjusts the input references in subsequent nodes.

    Parameters:
        model (onnx.ModelProto): The ONNX model to modify.

    Returns:
        onnx.ModelProto: The modified ONNX model with the cast operation added.

    Example:
        >>> import onnx
        >>> model = onnx.load("model.onnx")
        >>> modified_model = cast_double_to_float(model)
        >>> onnx.save(modified_model, "modified_model.onnx")
    """
    input_name = model.graph.input[0].name
    cast_node = onnx.helper.make_node(
        "Cast",
        inputs=[input_name],
        outputs=["cast_float_input"],
        to=onnx.TensorProto.FLOAT,  # Target type is float32
    )
    for node in model.graph.node:
        for i, input_name in enumerate(node.input):
            if input_name == model.graph.input[0].name:
                node.input[i] = "cast_float_input"

    model.graph.input[0].type.tensor_type.elem_type = onnx.TensorProto.DOUBLE
    model.graph.node.insert(0, cast_node)
    return model


class ONNXClient(ABC):
    _BINARY_DATA_OP = "binaryData()"
    _RESULT_TYPE_TO_MODEL_TYPE = {
        "OnnxPredictionModel": "PREDICTION",
        "OnnxAnomalyModel": "ANOMALY",
    }

    def _build_options(self, model_type: str, options: Dict[str, str]) -> str:
        allowed_options = {
            "PREDICTION": {"valueUnits"},
            "ANOMALY": {"outlierValue", "outlierPredicate"},
        }
        allowed_prefixes = {
            "ANOMALY": {"singleValue", "featureValue"},
        }

        if model_type not in allowed_options:
            raise ValueError(
                f"Invalid model type: '{model_type}'. Must be one of {list(allowed_options.keys())}."
            )

        valid_options = allowed_options[model_type]
        valid_prefixes = allowed_prefixes.get(model_type, set())
        invalid_keys = [
            k
            for k in options
            if k not in valid_options and not any(k.startswith(p) for p in valid_prefixes)
        ]
        if invalid_keys:
            raise ValueError(f"Invalid options for {model_type}: {invalid_keys}")

        formatted_options = [
            f'"{key}", "{value}"' for key, value in options.items() if value is not None
        ]
        return f"options({', '.join(formatted_options)})"

    def _generate_formula(self, model_type: str, labels: List[str], options: Dict[str, str]) -> str:

        operator_map = {
            "PREDICTION": "toPredictionModel",
            "ANOMALY": "toAnomalyModel",
        }

        if model_type not in operator_map:
            raise ValueError(
                f"Invalid model type: '{model_type}'. Must be one of {list(operator_map.keys())}."
            )

        operator = operator_map[model_type]
        options_str = self._build_options(model_type, options)
        return f"{self._BINARY_DATA_OP}.{operator}({len(labels)}, {options_str})"

    def _build_inputs(self, labels) -> List[Dict[str, str]]:
        value = json.dumps([{"name": label} for label in labels])
        prop = {"name": PropertyNames.Inputs, "value": value}
        return [prop]

    def _base64encoded_model(self, model: onnx.ModelProto) -> str:
        model_bytes = io.BytesIO()
        onnx.save(model, model_bytes)
        model_bytes.seek(0)
        return base64.b64encode(model_bytes.read()).decode("utf-8")

    def _list_formated(self, items) -> pd.DataFrame:
        records = []
        for item in items:
            base = {
                "ID": item.get("id"),
                "Name": item.get("name"),
                "Description": item.get("description"),
                "Archived": item.get("isArchived"),
            }
            props = item.get("includedProperties")
            inps = [inp["name"] for inp in json.loads(props.get(PropertyNames.Inputs, {}).get("value", "[]"))]
            base["Input Labels"] = inps
            base["Feature Count"] = len(inps)
            base["Type"] = self._RESULT_TYPE_TO_MODEL_TYPE[props.get(PropertyNames.ResultType, {}).get("value")]
            base["Scoped To"] = props.get(PropertyNames.ScopedTo, {}).get("value")
            records.append(base)

        column_order = ["ID", "Name", "Description", "Type", "Input Labels", "Feature Count", "Scoped To", "Archived"]

        return pd.DataFrame.from_records(records, columns=column_order)

    def register(self, name: str, description: str, model: onnx.ModelProto, model_type: str, labels: List[str],
                 scope: str = None, **kwargs) -> str:
        """
        Registers an ONNX in Seeq.

        Parameters:
            name: The name of the ONNX to register.
            description: The description of the ONNX.
            model: The ONNX to register. Must be a valid ONNX.
            model_type: The type of the ONNX. Must be one of "PREDICTION" or "ANOMALY".
            labels: The labels for the ONNX inputs. Must be provided as a list of strings.
            scope: The ID of the workbook to scope the ONNX to. If not provided, the ONNX will not be scoped.
            kwargs: Additional options for the ONNX, such as `valueUnits`, `outlierValue`, etc.

        Returns:
            ID of the registered ONNX.
        """
        raise NotImplementedError("Method 'register' must be implemented in subclasses.")

    def update(self, id: str, name: str = None, description: str = None, model: onnx.ModelProto = None,
               labels: List[str] = None, scope: str = None, **kwargs) -> None:
        """
        Updates an existing ONNX in Seeq.

        Parameters:
            id: The ID of the ONNX to update.
            name: The new name for the ONNX.
            description: The new description for the ONNX.
            model: The ONNX to update. If provided, the model must be a valid ONNX.
            labels: The labels for the ONNX inputs. Must be provided if `model` is provided.
            scope: The ID of the workbook to scope the ONNX to. If not provided, the ONNX scope will be left unchanged. Note: Globally scoped ONNX cannot be updated to a workbook scope.
            kwargs: Additional options for the ONNX, such as `valueUnits`, `outlierValue`, etc.

        Returns:
            None
        """
        raise NotImplementedError("Method 'update' must be implemented in subclasses.")

    def list(self, scope: str = None, include_archived: bool = False) -> pd.DataFrame:
        """
        Lists all ONNX models registered in Seeq.

        Parameters:
            scope: The ID of the workbook to filter the ONNX by. If not provided, all ONNX will be listed.
            include_archived: If True, includes archived ONNX in the list. Defaults to False.

        Returns:
            pd.DataFrame containing the ONNX with their details.
        """
        raise NotImplementedError("Method 'list' must be implemented in subclasses.")

    def archive(self, id: str, delete: bool = False) -> None:
        """
        Archives or deletes an ONNX by its ID.

        Parameters:
            id: The ID of the ONNX to archive or delete.
            delete: If True, the ONNX will be deleted; otherwise, it will be archived.
        Returns:
            None
        """
        raise NotImplementedError("Method 'archive' must be implemented in subclasses.")


def _custom_response_hook(response: requests.Response, *args, **kwargs) -> None:
    http_error_msg = ""
    reason = response.json().get("statusMessage", "Unknown error")
    if 400 <= response.status_code < 500:
        http_error_msg = f"{response.status_code} Client Error: {reason} for url: {response.url}"

    elif 500 <= response.status_code < 600:
        http_error_msg = f"{response.status_code} Server Error: {reason} for url: {response.url}"

    if http_error_msg:
        raise requests.exceptions.HTTPError(http_error_msg, response=response)


class SeeqONNXClient(ONNXClient):
    def __init__(self, host: str, auth_token: str):
        self._host = host
        self._auth_token = auth_token
        self._session = requests.Session()
        self._session.hooks = {'response': _custom_response_hook}
        self._session.headers.update({
            "Content-Type": "application/vnd.seeq.v1+json",
            "Accept": "application/vnd.seeq.v1+json",
            "x-sq-auth": self._auth_token})

    def list(self, scope: str = None, include_archived: bool = False) -> pd.DataFrame:
        url = urljoin(self._host, "api/items")
        filters = [f"{PropertyNames.ResultType}=={FormulaType.OnnxPredictionModel}",
                   f"{PropertyNames.ResultType}=={FormulaType.OnnxAnomalyModel}"]
        if include_archived:
            filters.append("@includeUnsearchable")
        params = {
            "includeProperties": [PropertyNames.Inputs, PropertyNames.ResultType, PropertyNames.ScopedTo],
            "filters": filters,
            "types": ["FormulaItem"]
        }
        if scope:
            params["scope"] = scope
        items = self._session.get(url, params=params)
        return self._list_formated(items.json().get("items", []))

    def register(self, name: str, description: str, model: onnx.ModelProto, model_type: str, labels: List[str],
                 scope: str = None, **kwargs) -> str:
        url = urljoin(self._host, "api/formulas/items")
        payload = {
            "name": name,
            "description": description,
            "formula": self._generate_formula(model_type, labels, kwargs),
            "additionalProperties": self._build_inputs(labels),
            "binaryData": self._base64encoded_model(model),
            "scopedTo": scope,
        }
        response = self._session.post(url, data=json.dumps(payload)).json()
        print(f"Successfully registered ONNX '{response['name']}' (ID: {response['id']}) (Scoped to: {scope})")
        return response['id']

    def update(self, id: str, name: str = None, description: str = None, model: onnx.ModelProto = None,
               labels: List[str] = None, scope: str = None, **kwargs) -> None:
        result_type_prop_url = urljoin(self._host, f"api/items/{id}/properties/{PropertyNames.ResultType}")
        result_type = self._session.get(result_type_prop_url).json().get("value")
        if result_type not in self._RESULT_TYPE_TO_MODEL_TYPE.keys():
            raise ValueError(f"Item with id {id} is not a valid ONNX. Result Type: {result_type}")
        model_type = self._RESULT_TYPE_TO_MODEL_TYPE[result_type]
        if model:
            if not labels:
                raise ValueError("Labels must be provided when updating the model.")
            set_formula_url = urljoin(self._host, f"api/items/{id}/formula")
            payload = {
                "binaryData": self._base64encoded_model(model),
                "formula": self._generate_formula(model_type, labels, kwargs)
            }
            self._session.post(set_formula_url, data=json.dumps(payload))
        properties = []
        if name:
            properties.append({"name": PropertyNames.Name, "value": name})
        if description:
            properties.append({"name": PropertyNames.Description, "value": description})
        if labels:
            properties.extend(self._build_inputs(labels))
        if properties:
            set_properties_url = urljoin(self._host, f"api/items/{id}/properties")
            self._session.post(set_properties_url, data=json.dumps(properties))
        if scope:
            set_scope_url = urljoin(self._host, f"api/items/{id}/scope")
            params = {"workbookId": scope}
            self._session.post(set_scope_url, params=params)
        print(f"Successfully updated ONNX (ID: {id})")

    def archive(self, id: str, delete: bool = False) -> None:
        url = urljoin(self._host, f"api/items/{id}")
        params = {
            "delete": False,
            "archivedReason": "BY_USER"
        }
        self._session.delete(url, params=params)
        if delete:
            params["delete"] = True
            self._session.delete(url, params=params)
            print(f"Successfully deleted ONNX (ID: {id})")
            return
        print(f"Successfully archived ONNX (ID: {id})")



from __future__ import annotations

import json
import os
import pickle
import types
from typing import List, Optional

import pandas as pd

from seeq import spy
from seeq.base import util
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks._context import WorkbookPushContext
from seeq.spy.workbooks._item import Item
from seeq.spy.workbooks._item_map import ItemMap
from seeq.spy.workbooks._workbook import Workbook
from seeq.spy.workbooks.job import _pull


@Status.handle_keyboard_interrupt()
def push(job_folder, *, resume: bool = True, path: str = None, owner: str = None, label: str = None, datasource=None,
         use_full_path: bool = False, access_control: str = None, override_max_interp: bool = False,
         global_inventory: Optional[str] = None, create_dummy_items: bool = False,
         errors: Optional[str] = None, quiet: Optional[bool] = None,
         status: Optional[Status] = None, session: Optional[Session] = None,
         scope_globals_to_workbook: Optional[bool] = None) -> pd.DataFrame:
    """
    Pushes the definitions for each workbook that was pulled by the
    spy.workbooks.job.pull() function, in a restartable "job"-like fashion.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder created by
        spy.workbooks.job.pull().

    resume : bool, default True
        True if the push should resume from where it left off, False if it
        should push everything again.

    path : str, default None
        A '>>'-delimited folder path to create to contain the workbooks. Note
        that a further subfolder hierarchy will be created to preserve the
        relative paths that the folders were in when they were searched for
        and pulled. If you specify None, then the workbook will stay where
        it is (if it has already been pushed once).

        If you specify spy.workbooks.MY_FOLDER, it will be moved to the user's
        home folder.

        If you specify a folder ID directly, it will be pushed to that folder.

        If you specify spy.workbooks.ORIGINAL_FOLDER, it will be pushed to the
        folder it was in when it was originally pulled. The folder hierarchy
        will be recreated on the target server if it doesn't already exist.
        (You must be an admin to use this to ensure you have permissions to
        put things where they need to go.)

    owner : str, default None
        Determines the ownership of pushed workbooks and folders.

        By default, the current owner will be preserved. If the content doesn't
        exist yet, the logged-in user will be the owner.

        All other options require that the logged-in user is an admin:

        If spy.workbooks.ORIGINAL_OWNER, ownership is assigned according to the
        original owner of the pulled content.

        If spy.workbooks.FORCE_ME_AS_OWNER, existing content will be
        reassigned to the logged-in user.

        If a username or a user's Seeq ID is supplied, that user will be
        assigned as owner.

        You may need to supply an appropriate datasource map if the usernames
        are different between the original and the target servers.

    label : str
        A user-defined label that differentiates this push operation from
        others. By default, the label will be the logged-in user's username
        OR the username from the 'owner' argument so that push activity will
        generally be isolated by user. But you can override this with a label
        of your choosing.

    datasource : str, optional, default 'Seeq Data Lab'
        The name of the datasource within which to contain all the pushed items.
        Items inherit access control permissions from their datasource unless it
        has been overridden at a lower level. If you specify a datasource using
        this argument, you can later manage access control (using spy.acl functions)
        at the datasource level for all the items you have pushed.

        If you instead want access control for your items to be inherited from the
        workbook they are scoped to, specify `spy.INHERIT_FROM_WORKBOOK`.

    use_full_path : bool, default False
        If True, the original full path for an item is reconstructed, as
        opposed to the path that is relative to the Path property supplied to
        the spy.workbooks.search() call that originally helped create these
        workbook definitions. Note that this full path will still be inside
        the folder specified by the 'path' argument, if supplied.

    access_control : str, default None
        Specifies how Access Control Lists should be treated, via the
        following keywords: add/replace,loose/strict

        - If None, then no access control entries are pushed.
        - If 'add', then existing access control entries will not be disturbed
          but new entries will be added.
        - If 'replace', then existing access control entries will be removed
          and replaced with the entries from workbook definitions.
        - If 'loose', then any unmapped users/groups from the workbook
          definitions will be silently ignored.
        - If 'strict', then any unmapped users/groups will result in errors.

        Example: access_control='replace,loose'

    override_max_interp : bool, default False
        If True, then the Maximum Interpolation overrides from the source
        system will be written to the destination system.

    global_inventory : {'copy global', copy local', 'always reuse'}, default 'copy global'
        Determines how SPy handles global inventory items, especially in
        conjunction with a label argument. 'copy global' will cause different
        global items to be created if their labels differ. If no label is
        specified, the existing global item will be reused/updated if possible.
        'copy local' will scope the global items to the pushed workbook,
        making copies for different workbooks if necessary. 'always reuse' will
        use an existing global item if possible and ignore the label.

    create_dummy_items : bool, default False
        If true, then "dummy" items will be created for any stored items
        that are not successfully mapped to the target system. This should
        be specified as True if you intend to use the spy.workbooks.job.data
        module. The dummy items will be created in the target system's
        "Seeq Data Lab" datasource and will have the same name as the original
        item, and no data. They will be scoped to a workbook under a
        "SPy Workbook Jobs" folder and named after the job folder.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame (errors='catalog' must be combined with
        status=<Status object>).

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    scope_globals_to_workbook : bool, default False
        Deprecated. Use global_inventory instead: 'copy global' is the
        equivalent of False, 'copy local' is the equivalent of True.

    """
    input_args = _common.validate_argument_types([
        (job_folder, 'job_folder', str),
        (resume, 'resume', bool),
        (path, 'path', str),
        (owner, 'owner', str),
        (label, 'label', str),
        (datasource, 'datasource', str),
        (use_full_path, 'use_full_path', bool),
        (access_control, 'access_control', str),
        (override_max_interp, 'override_max_interp', bool),
        (global_inventory, 'global_inventory', str),
        (create_dummy_items, 'create_dummy_items', bool),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session),
        (scope_globals_to_workbook, 'scope_globals_to_workbook', bool)
    ])

    _login.validate_login(session, status)

    if not util.safe_exists(job_folder):
        raise SPyValueError(f'Job folder "{job_folder}" does not exist.')

    item_map = ItemMap()
    if resume:
        item_map = load_item_map(job_folder)

    item_map.only_override_maps = True

    workbook_list: List[WorkbookFolderRef] = list()
    job_workbooks_folder = _pull.get_workbooks_folder(job_folder)
    for workbook_folder, _ in _pull.walk_workbook_folders(job_workbooks_folder):
        completely_pushed_filename = os.path.join(job_workbooks_folder, workbook_folder, 'Completely Pushed')
        if not resume and util.safe_exists(completely_pushed_filename):
            util.safe_remove(completely_pushed_filename)
        workbook_list.append(WorkbookFolderRef(os.path.join(job_workbooks_folder, workbook_folder), job_folder))

    job_datasource_maps_folder = _pull.get_datasource_maps_folder(job_folder)
    if not util.safe_exists(job_datasource_maps_folder):
        job_datasource_maps_folder = None

    create_dummy_items_in_workbook = get_dummy_workbook_name(job_folder) if create_dummy_items else None

    spy.workbooks.push(workbook_list, path=path, owner=owner, label=label, datasource=datasource,
                       use_full_path=use_full_path, access_control=access_control,
                       override_max_interp=override_max_interp, global_inventory=global_inventory,
                       datasource_map_folder=job_datasource_maps_folder, refresh=False,
                       create_dummy_items_in_workbook=create_dummy_items_in_workbook,
                       status=status, session=session, item_map=item_map,
                       scope_globals_to_workbook=scope_globals_to_workbook)

    results_df = status.df.copy()

    results_df_properties = types.SimpleNamespace(
        func='spy.workbooks.job.push',
        kwargs=input_args,
        status=status,
        item_map=item_map
    )

    _common.put_properties_on_df(results_df, results_df_properties)

    return results_df


def get_item_map_filename(job_folder):
    return os.path.join(job_folder, 'item_map.pickle')


def load_item_map(job_folder):
    item_map_file = get_item_map_filename(job_folder)
    if not util.safe_exists(item_map_file):
        return ItemMap()

    with util.safe_open(item_map_file, 'rb') as f:
        return pickle.load(f)


def save_item_map(job_folder, item_map):
    item_map_file = get_item_map_filename(job_folder)
    with util.safe_open(item_map_file, 'wb') as f:
        pickle.dump(item_map, f, protocol=_common.DEFAULT_PICKLE_PROTOCOL)


def get_dummy_workbook_name(job_folder):
    return f'SPy Dummy Workbooks >> Dummy Workbook for {job_folder}'


def redo(job_folder: str, status: Status):
    job_workbooks_folder = _pull.get_workbooks_folder(job_folder)
    workbook_folders = {i: f for f, i in _pull.walk_workbook_folders(job_workbooks_folder)}
    workbook_ids: pd.Series = status.df['ID']
    for index, workbook_id in workbook_ids.items():
        if workbook_id in workbook_folders:
            completely_pushed_filename = os.path.join(
                job_workbooks_folder, workbook_folders[workbook_id], 'Completely Pushed')
            if util.safe_exists(completely_pushed_filename):
                util.safe_remove(completely_pushed_filename)
            status.df.at[index, 'Result'] = 'Push will be redone'
        else:
            status.df.at[index, 'Result'] = 'Not found'


class WorkbookFolderRef(Workbook):
    workbook_folder: str
    _real_workbook: Optional[Workbook]
    _parent_folder_id: Optional[str]

    def __init__(self, workbook_folder, job_folder):
        super().__init__()

        self.workbook_folder = workbook_folder
        self.job_folder = job_folder
        self._provenance = Item.LOAD
        self._real_workbook = None
        self._parent_folder_id = None

        with util.safe_open(os.path.join(self.workbook_folder, 'Workbook.json'), 'r', encoding='utf-8') as f:
            self._definition = json.load(f)

    def _already_pushed_filename(self):
        return os.path.join(self.workbook_folder, 'Completely Pushed')

    @property
    def already_pushed(self):
        return util.safe_exists(self._already_pushed_filename())

    def _ensure_loaded(self):
        if self._real_workbook is not None:
            return

        self._real_workbook = Workbook.load(self.workbook_folder)
        self._real_workbook.datasource_maps = self.datasource_maps

    def push_containing_folders(self, session: Session, item_map: ItemMap, datasource_output, use_full_path,
                                parent_folder_id, owner, label, access_control, status: Status):
        if self.already_pushed:
            with util.safe_open(self._already_pushed_filename(), 'r') as f:
                d = json.load(f)
            self._parent_folder_id = d['Parent Folder ID']
            return self._parent_folder_id

        self._ensure_loaded()
        self._parent_folder_id = self._real_workbook.push_containing_folders(
            session, item_map, datasource_output, use_full_path, parent_folder_id, owner, label, access_control, status)

        return self._parent_folder_id

    def push(self, *, context: WorkbookPushContext, folder_id=None, item_map: ItemMap = None, label=None,
             include_inventory=True):
        if self.already_pushed:
            self._push_errors = Workbook.load_push_errors(self.workbook_folder)
            context.status.put('Pushed Workbook ID', item_map[self.id])
            link_url = Workbook.construct_url(
                context.session,
                folder_id,
                item_map[self.id]
            )
            context.status.put('URL', link_url)
            return

        self._real_workbook.push(context=context, folder_id=folder_id, item_map=item_map, label=label,
                                 include_inventory=include_inventory)

        save_item_map(self.job_folder, item_map)
        self._push_errors = self._real_workbook.push_errors
        Workbook.save_push_errors(self.workbook_folder, self._push_errors)
        if context.status.errors == 'catalog' or len(self._real_workbook.push_errors) == 0:
            with util.safe_open(self._already_pushed_filename(), 'w') as f:
                json.dump({'Parent Folder ID': self._parent_folder_id}, f)



from __future__ import annotations

"""
This module houses all the shared functionality in the build lib that is used among multiple modules. Try to use these
functions wherever possible.
"""
from datetime import timedelta, tzinfo

#
# See https://stackoverflow.com/a/2331635/2452569 for why this is necessary
#
ZERO = timedelta(0)


class UTC(tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO


utc = UTC()


# See http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named
class Bunch:

    def __init__(self, **kwds):
        self.__dict__.update(kwds)



from __future__ import annotations

import copy
import re
import types
from typing import Optional

import numpy as np
import pandas as pd

from seeq.sdk import *
from seeq.spy import _common, _login, _metadata, _redaction
from seeq.spy._errors import *
from seeq.spy._redaction import safely
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.assets._model import ItemGroup
from seeq.spy.assets._trees import _constants, _match, _path, _utils
from seeq.spy.assets._trees._pandas import KeyedDataFrame


def apply_friendly_name(df) -> None:
    if 'Friendly Name' not in df.columns or df['Friendly Name'].isnull().all():
        _common.put_properties_on_df(df, types.SimpleNamespace(modified_items=set()))
        return

    # If we are changing the names of items in a dataframe whose paths are dependent on one another, then
    # record those dependencies so we can modify paths afterwards as well
    relationships = path_relationships(df)

    modified_items = set()
    for i in df.index:
        if pd.isnull(df.loc[i, 'Friendly Name']):
            continue
        if _match.is_column_value_query(df.loc[i, 'Friendly Name']):
            new_name = _match.fill_column_values(df.loc[i], df.loc[i, 'Friendly Name'])
        else:
            new_name = df.loc[i, 'Friendly Name']
        if pd.isnull(new_name):
            continue
        df.loc[i, 'Name'] = new_name
        if _common.present(df.loc[i], 'ID'):
            modified_items.add(df.loc[i, 'ID'])

    recover_relationships(df, relationships)
    _common.put_properties_on_df(df, types.SimpleNamespace(modified_items=modified_items))


class _PathRelationshipNode:
    """
    Helper class for path_relationships()
    """

    def __init__(self, name, index=None):
        self.name = name
        self.index = index
        self.children = dict()

    @staticmethod
    def tree_root() -> _PathRelationshipNode:
        return _PathRelationshipNode(None)

    def add_branch(self, path_list, index, offset=0):
        if offset >= len(path_list):
            return
        include_index = offset == len(path_list) - 1
        name = path_list[offset].casefold()
        if name not in self.children:
            self.children[name] = _PathRelationshipNode(name)
        self.children[name].add_branch(path_list, index, offset + 1)
        if include_index:
            self.children[name].index = index

    def calculate_relationships(self, relationship_dict=None, branch_references=None, offset=0) -> dict:
        if relationship_dict is None:
            relationship_dict = dict()
        if branch_references is None:
            branch_references = dict()
        for node in self.children.values():
            if node.index is not None:
                if len(branch_references) > 0:
                    relationship_dict[node.index] = branch_references.copy()
                branch_references[offset] = node.index
            node.calculate_relationships(relationship_dict, branch_references, offset + 1)
            branch_references.pop(offset, None)
        return relationship_dict


def path_relationships(df) -> Optional[dict]:
    """
    Return a dict of dicts indicating via integers how the paths of the input rows are dependent on one another.

    Example:
        df = pd.DataFrame([{
            'Path': 'Root', 'Name': 'Item 1'
        }, {
            'Path': 'Root >> Item 1', 'Name': 'Item 2'
        }])

        output = {
            1: { # 1 refers here to the row in df with index 1, i.e., Item 2
                1: 0 # 1 refers here to the item in Item 2's path with index 1, i.e. 'Item 1'
                     # 0 refers here to the index of Item 1's row in df
            }
        }
    """
    if len(df) == 0 or 'Name' not in df.columns or 'Path' not in df.columns:
        return None
    full_paths = list(df.apply(_path.get_full_path, axis=1, check_asset_column=True).apply(_common.path_string_to_list))
    path_relationship_tree = _PathRelationshipNode.tree_root()
    for i, full_path in enumerate(full_paths):
        path_relationship_tree.add_branch(full_path, i)
    return path_relationship_tree.calculate_relationships()


def recover_relationships(df, relationships) -> None:
    """
    Takes a list of relationships (in the format described in _path_relationships) and modifies paths in
    df to reflect those relationships
    """
    if relationships is None:
        return
    for i, path_ref_dict in relationships.items():
        path = _path.determine_path(df.loc[i])
        path_list = _common.path_string_to_list(path) if path else []
        for j, reference in path_ref_dict.items():
            if 0 <= reference < len(df) and 0 <= j < len(path_list):
                path_list[j] = df.loc[reference, 'Name']
        df.loc[i, 'Path'] = _common.path_list_to_string(path_list)
    if 'Asset' in df.columns:
        df.drop(columns='Asset', inplace=True)


def process_properties(session: Session, df, display_template_map, datasource, status: Status, *, existing_tree_df=None,
                       pull_nodes=True, keep_parent_column=False) -> KeyedDataFrame:
    """
    Sanitize and pull item properties into an input dataframe. Steps in order:
    -- Pulls missing properties for items with ID provided
    -- Filters out properties not in _constants.dataframe_columns
    -- Determines tree depth
    -- Determines (if possible_tree_copy is True) if the input dataframe corresponds to an existing SPy tree
        -- If it is indeed a copy of a SPy tree, pulls in calculations from the original tree
        -- Otherwise, it converts all items with IDs into references
    -- Ensures all formula parameters are NAN or dict
    """
    df = df.reset_index(drop=True)
    df = KeyedDataFrame.of(df)

    _safely = _redaction.request_safely(
        action_description='get properties for requested item',
        status=status,
        default_value=pd.Series(np.nan, index=_constants.dataframe_columns, dtype=object))

    df = df.apply(_safely(process_row_properties), axis=1,
                  session=session,
                  status=status,
                  pull_nodes=pull_nodes,
                  keep_parent_column=keep_parent_column)

    datasource = datasource if datasource is not None else _common.DEFAULT_DATASOURCE_ID

    def _row_is_from_existing_tree(row):
        if existing_tree_df is None or not _common.present(row, 'ID'):
            return 'new'
        same_id_rows = existing_tree_df[existing_tree_df.ID.str.casefold() == row['ID'].casefold()]
        if len(same_id_rows) != 1:
            return 'new'
        if _common.present(row, 'Type') and row['Type'].casefold() != same_id_rows.Type.iloc[0].casefold():
            return 'new'
        if _common.present(row, 'Datasource Class') and row['Datasource Class'] != _common.DEFAULT_DATASOURCE_CLASS:
            return 'new'
        if _common.present(row, 'Datasource ID') and row['Datasource ID'] != datasource:
            return 'new'
        if _common.present(row, 'Name') and row['Name'].casefold() != same_id_rows.Name.iloc[0].casefold():
            return 'modified'
        if _common.present(row, 'Path') and row['Path'].casefold() != same_id_rows.Path.iloc[0].casefold():
            return 'modified'
        return 'pre-existing'

    row_type = df.apply(_row_is_from_existing_tree, axis=1)
    df.drop(columns=['Datasource Class', 'Datasource ID'], inplace=True, errors='ignore')
    modified_items = df.loc[row_type == 'modified', 'ID'] if 'ID' in df.columns else set()

    # For the nodes that originated from the pre-existing SPy tree we are modifying, we want to pull
    # pre-existing calculations directly.
    df.loc[row_type == 'pre-existing', :] = df.loc[row_type == 'pre-existing', :] \
        .apply(_safely(pull_calculation), axis=1, session=session)

    # For the nodes that originate from places other than the pre-existing SPy tree we are modifying,
    # we want to build references so we create and modify *copies* and not the original items.
    df.loc[row_type != 'pre-existing', :] = df.loc[row_type != 'pre-existing', :] \
        .apply(_safely(make_node_reference), axis=1, display_template_map=display_template_map,
               session=session)

    if 'Formula Parameters' in df.columns:
        df['Formula Parameters'] = df['Formula Parameters'].apply(formula_parameters_to_dict)

    # Drop all-NaN rows created by safely() calls
    # TODO CRAB-41833: KeyedDataFrame.dropna doesn't work as expected.
    df = KeyedDataFrame.of(pd.DataFrame(df).dropna(how='all'))
    if 'Depth' in df:
        df['Depth'] = df['Depth'].astype('int32')

    _common.put_properties_on_df(df, types.SimpleNamespace(modified_items=modified_items))

    return df


# Note that the session argument is second in this special case because we call this from pd.DataFrame.apply(),
# which requires that row is the first argument.
def process_row_properties(row, session: Session, status, pull_nodes, keep_parent_column) -> pd.Series:
    if _common.present(row, 'ID') and pull_nodes:
        new_row = pull_node(session, row['ID'])
        _utils.increment_status_df(status, pulled_items=[new_row])
    else:
        new_row = pd.Series(index=_constants.dataframe_columns, dtype=object)

    # In case that properties are specified, but IDs are given, the user-given properties
    # override those pulled from Seeq
    for prop, value in row.items():
        if prop == 'Type' and _common.present(new_row, 'Type'):
            continue
        if prop in ['Path', 'Asset']:
            prop = 'Path'
            value = _path.determine_path(row)
        add_tree_property(new_row, prop, value)

    if not _common.present(new_row, 'Type') and not _common.present(new_row, 'Formula') \
            and not _common.present(new_row, 'Roll Up Statistic'):
        if _common.present(new_row, 'Measured Item'):
            new_row['Type'] = 'ThresholdMetric'
        else:
            new_row['Type'] = 'Asset'

    if not _common.present(new_row, 'Path'):
        new_row['Path'] = ''
    new_row['Depth'] = new_row['Path'].count('>>') + 2 if new_row['Path'] else 1

    if keep_parent_column and _common.present(row, 'Parent'):
        new_row['Parent'] = row['Parent']

    return new_row


def make_node_reference(row, display_template_map, session: Session) -> pd.Series:
    row = row.copy()
    if _common.present(row, 'ID'):
        if _common.get(row, 'Type') in _constants.data_types and not is_reference(row):
            _metadata.build_reference(session, row)
        elif 'Metric' in _common.get(row, 'Type'):
            row = pull_calculation(row, session)
        elif _common.get(row, 'Type') == 'Display':
            row = row.copy()
            displays_api = DisplaysApi(session.client)
            template = displays_api.get_display(id=row['ID']).template  # type: DisplayTemplateOutputV1
            template_id = template.id.upper()
            row['Template ID'] = template_id

            if template_id not in display_template_map:
                display_template_map[template_id] = DisplayTemplateInputV1(
                    name=template.name,
                    description=template.description,
                    source_workstep_id=template.source_workstep_id,
                    swap_source_asset_id=template.swap_source_asset_id
                )
        if _common.present(row, 'ID'):
            row['Referenced ID'] = row['ID']
    row['ID'] = np.nan
    return row


def is_reference(row) -> bool:
    if not _common.get(row, 'Referenced ID') or not _common.get(row, 'Formula Parameters'):
        return False
    formula = _common.get(row, 'Formula')
    if formula is not None and re.match(r'^\$\w+$', formula):
        return True
    else:
        return False


def pull_calculation(row, session: Session) -> pd.Series:
    formulas_api = FormulasApi(session.client)
    metrics_api = MetricsApi(session.client)
    displays_api = DisplaysApi(session.client)
    if _common.get(row, 'Type') in _constants.calculated_types and _common.present(row, 'ID'):
        row = row.copy()
        formula_output = formulas_api.get_item(id=row['ID'])  # type: FormulaItemOutputV1
        row['Formula'] = formula_output.formula
        row['Formula Parameters'] = {
            p.name: p.item.id if p.item else p.formula
            for p in formula_output.parameters
        }
    elif _common.get(row, 'Type') == 'Display' and _common.present(row, 'ID'):
        row = row.copy()
        display_output = displays_api.get_display(id=row['ID'])
        row['Template ID'] = display_output.template.id.upper()
    elif 'Metric' in _common.get(row, 'Type') and _common.present(row, 'ID'):
        row = row.copy()
        row['Type'] = 'ThresholdMetric'

        def str_from_scalar_value_output(scalar_value_output):
            # noinspection PyProtectedMember
            return _metadata.str_from_scalar_value_dict(
                _metadata.dict_from_scalar_value_output(scalar_value_output)).strip()

        def set_metric_value_if_present(column_name, val):
            if val is not None:
                if isinstance(val, ScalarValueOutputV1):
                    row[column_name] = str_from_scalar_value_output(val)
                elif isinstance(val, ItemPreviewWithAssetsV1):
                    row[column_name] = val.id
                else:
                    row[column_name] = val

        metric = metrics_api.get_metric(id=row['ID'])  # type: ThresholdMetricOutputV1
        set_metric_value_if_present('Aggregation Function', metric.aggregation_function)
        set_metric_value_if_present('Bounding Condition', metric.bounding_condition)
        set_metric_value_if_present('Bounding Condition Maximum Duration', metric.bounding_condition_maximum_duration)
        set_metric_value_if_present('Duration', metric.duration)
        set_metric_value_if_present('Measured Item', metric.measured_item)
        set_metric_value_if_present('Metric Neutral Color', metric.neutral_color)
        set_metric_value_if_present('Number Format', metric.number_format)
        set_metric_value_if_present('Period', metric.period)
        set_metric_value_if_present('Process Type', metric.process_type)
        if metric.thresholds:
            thresholds_dict = dict()
            for threshold in metric.thresholds:  # type: ThresholdOutputV1
                # Key can be 'HiHi#FF0000' or 'HiHi'
                key = threshold.priority.name
                if threshold.priority.color:
                    key += threshold.priority.color
                # Value could get pulled as an ID or as a string-ified value
                value = ''
                if not threshold.is_generated and threshold.item:
                    value = threshold.item.id
                elif threshold.value is not None:
                    if isinstance(threshold.value, ScalarValueOutputV1):
                        value = str_from_scalar_value_output(threshold.value)
                    else:
                        value = threshold.value
                if key and value:
                    thresholds_dict[key] = value
            row['Thresholds'] = thresholds_dict

    return row


def pull_node(session: Session, node_id) -> pd.Series:
    """
    Returns a dataframe row corresponding to the item given by node_id
    """
    items_api = _login.get_api(session, ItemsApi)

    item_output = items_api.get_item_and_all_properties(id=node_id)  # type: ItemOutputV1
    node = pd.Series(index=_constants.dataframe_columns, dtype=object)

    # Extract only the properties we use
    node['Name'] = item_output.name
    node['Type'] = item_output.type
    node['ID'] = item_output.id  # If this should be a copy, it'll be converted to 'Referenced ID' later
    for prop in item_output.properties:  # type: PropertyOutputV1
        add_tree_property(node, prop.name, prop.value)
    if 'Metric' in node['Type']:
        node = pull_calculation(node, session)

    return node


def add_tree_property(properties, key, value) -> pd.Series:
    """
    If the property is one which is used by SPy Trees, adds the key+value pair to the dict.
    """
    if key in _constants.dataframe_columns or key in ['Datasource Class', 'Datasource ID']:
        value = _common.none_to_nan(value)
        if key not in properties or not (pd.api.types.is_scalar(value) and pd.isnull(value)):
            properties[key] = value
    return properties


def formula_parameters_to_dict(formula_parameters) -> dict:
    if isinstance(formula_parameters, dict) or (pd.api.types.is_scalar(formula_parameters) and pd.isnull(
            formula_parameters)):
        return formula_parameters

    if isinstance(formula_parameters, str):  # formula_parameters == 'x=2b17adfd-3308-4c03-bdfb-bf4419bf7b3a'
        # handle an empty string case
        if len(formula_parameters) == 0:
            return dict()
        else:
            formula_parameters = [formula_parameters]

    if isinstance(formula_parameters, pd.Series):
        formula_parameters = formula_parameters.tolist()

    formula_dictionary = dict()
    if isinstance(formula_parameters, list):  # formula_parameters == ['x=2b17adfd-3308-4c03-bdfb-bf4419bf7b3a', ...]
        for param in formula_parameters:  # type: str
            split_list = param.split('=')  # ['x', '2b17...']
            if len(split_list) != 2:
                raise SPyException(f'Formula Parameter: {param} needs to be in the format \'paramName=inputItem\'.')
            formula_dictionary[split_list[0].strip()] = split_list[1].strip()
    return formula_dictionary  # output == {'x': '2b17adfd-3308-4c03-bdfb-bf4419bf7b3a'}


def format_references(df) -> pd.DataFrame:
    out_df = df.copy()
    out_df['Formula Parameters'] = pd.Series(np.nan, index=df.index, dtype=object)
    out_df['Thresholds'] = pd.Series(np.nan, index=df.index, dtype=object)

    for node in _match.TreeNode.of(df):  # type: _match.TreeNode
        if pd.api.types.is_dict_like(df.loc[node.index, 'Formula Parameters']):
            formula_parameters = copy.deepcopy(df.loc[node.index, 'Formula Parameters'])
            for name, reference in formula_parameters.items():
                if not isinstance(reference, str) or _common.is_guid(reference):
                    continue
                formula_parameters[name] = node.resolve_reference(reference).full_path
            out_df.at[node.index, 'Formula Parameters'] = formula_parameters

        if isinstance((df.loc[node.index, 'Measured Item']), str):
            if not _common.is_guid(df.loc[node.index, 'Measured Item']):
                out_df.at[node.index, 'Measured Item'] = node.resolve_reference(
                    df.loc[node.index, 'Measured Item']).full_path

        if isinstance(df.loc[node.index, 'Bounding Condition'], str):
            if not _common.is_guid(df.loc[node.index, 'Bounding Condition']):
                out_df.at[node.index, 'Bounding Condition'] = node.resolve_reference(
                    df.loc[node.index, 'Bounding Condition']).full_path

        if pd.api.types.is_dict_like(df.loc[node.index, 'Thresholds']):
            thresholds = copy.deepcopy(df.loc[node.index, 'Thresholds'])
            for name, reference in thresholds.items():
                if not isinstance(reference, str) or _common.is_guid(reference):
                    continue
                try:
                    path_list = _common.path_string_to_list(node.resolve_reference(reference).full_path)
                    # Use a dict so _push can differentiate a string Threshold and a reference to another item
                    thresholds[name] = {
                        'Path': _common.path_list_to_string(path_list[:-1]),
                        'Name': path_list[-1]
                    }
                except SPyRuntimeError as e:
                    # If reference was not found, we treat it like a string threshold and ignore
                    if 'invalid, missing, or has been removed' not in e.args[0]:
                        raise
            out_df.at[node.index, 'Thresholds'] = thresholds

        if isinstance(df.loc[node.index, 'Roll Up Parameters'], str):
            statistic = df.loc[node.index, 'Roll Up Statistic'].casefold()
            fn = next((fn for fn in _common.ROLL_UP_FUNCTIONS if fn.statistic == statistic), None)
            if fn is None:
                valid_statistics = set(fn.statistic.capitalize() for fn in _common.ROLL_UP_FUNCTIONS)
                raise SPyRuntimeError(f"Roll up statistic '{statistic}' not found. Valid options: "
                                      f"{', '.join(valid_statistics)}")
            items = node.resolve_references(df.loc[node.index, 'Roll Up Parameters'])
            formula_parameters = ItemGroup([item.full_path for item in items]).as_parameters()
            formula = fn.generate_formula(formula_parameters)
            out_df.at[node.index, 'Formula'] = formula
            out_df.at[node.index, 'Formula Parameters'] = formula_parameters

    return out_df


def push_and_replace_display_templates(session: Session, df, display_template_map, datasource, workbook_id,
                                       status: Status):
    display_templates_api = DisplayTemplatesApi(session.client)
    template_id_map = dict()
    datasource_output = _metadata.create_datasource(session, datasource)
    for original_id, display_template_input in display_template_map.items():
        display_template_input.datasource_class = datasource_output.datasource_class
        display_template_input.datasource_id = datasource_output.datasource_id
        display_template_input.scoped_to = workbook_id
        display_template_output = safely(
            lambda: display_templates_api.create_display_template(body=display_template_input),
            action_description=f'create new Display Template to replace original {original_id}',
            status=status)
        template_id_map[original_id] = display_template_output.id if display_template_output else None
    df['Template ID'] = df['Template ID'].replace(template_id_map)


def archive_and_remove_displays(session: Session, display_ids):
    displays_api = DisplaysApi(session.client)
    trees_api = TreesApi(session.client)
    for display_id in display_ids:
        try:
            trees_api.remove_node_from_tree(id=display_id)
        except ApiException:
            pass
        try:
            displays_api.archive_display(id=display_id)
        except ApiException:
            pass



from __future__ import annotations

import glob
import os
import tempfile
import zipfile
from typing import Optional

from seeq.base import util
from seeq.spy import _common
from seeq.spy._errors import *
from seeq.spy._status import Status
from seeq.spy.workbooks._template import package_as_templates
from seeq.spy.workbooks._workbook import Workbook, WorkbookList


@Status.handle_keyboard_interrupt(no_session=True)
def load(folder_or_zipfile, *, as_template_with_label: str = None, errors: Optional[str] = None,
         quiet: Optional[bool] = None, status: Optional[Status] = None) -> Optional[WorkbookList]:
    """
    Loads a list of workbooks from a folder on disk into Workbook objects in
    memory.

    Parameters
    ----------
    folder_or_zipfile : str
        A folder or zip file on disk containing workbooks to be loaded. Note
        that any subfolder structure will work -- this function will scan for
        any subfolders that contain a Workbook.json file and assume they should
        be loaded.

    as_template_with_label : str
        Causes the workbooks to be loaded as templates (either AnalysisTemplate
        or TopicTemplate) with the label specified. See the Workbook Templates
        documentation notebook for more information about templates.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame (errors='catalog' must be combined with
        status=<Status object>).

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed.
    """
    _common.validate_argument_types([
        (folder_or_zipfile, 'folder_or_zipfile', str),
        (as_template_with_label, 'as_template_with_label', str),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status)
    ])

    folder_or_zipfile = os.path.normpath(folder_or_zipfile)

    try:
        if not util.safe_exists(folder_or_zipfile):
            raise SPyRuntimeError('Folder/zipfile "%s" does not exist' % folder_or_zipfile)

        if folder_or_zipfile.lower().endswith('.zip'):
            with tempfile.TemporaryDirectory() as temp:
                with zipfile.ZipFile(util.handle_long_filenames(folder_or_zipfile), "r") as z:
                    status.update('Unzipping "%s"' % folder_or_zipfile, Status.RUNNING)
                    z.extractall(temp)

                status.update('Loading from "%s"' % temp, Status.RUNNING)
                workbooks = _load_from_folder(temp)
        else:
            status.update('Loading from "%s"' % folder_or_zipfile, Status.RUNNING)
            workbooks = _load_from_folder(folder_or_zipfile)

        if as_template_with_label is not None:
            workbooks = package_as_templates(workbooks, as_template_with_label)

        status.update('Success', Status.SUCCESS)
        return workbooks

    except KeyboardInterrupt:
        status.update('Load canceled', Status.CANCELED)
        return None


def _load_from_folder(folder):
    folder_escaped = glob.escape(folder)
    workbook_json_files = util.safe_glob(os.path.join(folder_escaped, '**', 'Workbook.json'), recursive=True)

    workbooks = WorkbookList()
    for workbook_json_file in workbook_json_files:
        workbooks.append(Workbook.load(os.path.dirname(workbook_json_file)))

    return workbooks
